from arcade_hubspot.enums import HubspotObject

HUBSPOT_OBJECT_PROPERTIES = {
    HubspotObject.COMPANY: [
        "type",
        "name",
        "about_us",
        "address",
        "city",
        "state",
        "zip",
        "country",
        "annualrevenue",
        "hs_annual_revenue_currency_code",
        "industry",
        "phone",
        "website",
        "domain",
        "numberofemployees",
        "hs_lead_status",
        "lifecyclestage",
        "linkedin_company_page",
        "twitterhandle",
    ],
    HubspotObject.CONTACT: [
        "salutation",
        "email",
        "work_email",
        "hs_additional_emails",
        "mobilephone",
        "phone",
        "firstname",
        "lastname",
        "lifecyclestage",
        "annualrevenue",
        "address",
        "date_of_birth",
        "degree",
        "gender",
        "buying_role",
        "hs_language",
        "hs_lead_status",
        "hs_timezone",
        "hubspot_owner_id",
        "hubspot_owner_name",
        "job_function",
        "jobtitle",
        "seniority",
        "industry",
        "twitterhandle",
    ],
    HubspotObject.CALL: [
        "hs_body_preview",
        "hs_call_direction",
        "hs_call_status",
        "hs_call_summary",
        "hs_call_title",
        "hs_createdate",
        "hubspot_owner_id",
        "hs_call_disposition",
        "hs_timestamp",
    ],
    HubspotObject.DEAL: [
        "dealname",
        "dealstage",
        "dealtype",
        "closedate",
        "closed_lost_reason",
        "closed_won_reason",
        "hs_is_closed_won",
        "hs_is_closed_lost",
        "hs_closed_amount",
        "pipeline",
        "hubspot_owner_id",
        "description",
        "hs_deal_score",
        "amount",
        "hs_forecast_probability",
        "hs_forecast_amount",
    ],
    HubspotObject.EMAIL: [
        "hs_object_id",
        "hs_body_preview",
        "hs_email_sender_raw",
        "hs_email_from_raw",
        "hs_email_to_raw",
        "hs_email_bcc_raw",
        "hs_email_cc_raw",
        "hs_email_subject",
        "hs_email_status",
        "hs_timestamp",
        "hubspot_owner_id",
        "hs_email_associated_contact_id",
    ],
    HubspotObject.MEETING: [
        "hs_object_id",
        "hs_meeting_title",
        "hs_body_preview",
        "hs_meeting_location",
        "hs_meeting_outcome",
        "hubspot_owner_id",
        "hs_meeting_start_time",
        "hs_meeting_end_time",
    ],
    HubspotObject.NOTE: [
        "hs_object_id",
        "hs_body_preview",
        "hs_timestamp",
        "hubspot_owner_id",
    ],
    HubspotObject.TASK: [
        "hs_object_id",
        "hs_body_preview",
        "hs_timestamp",
        "hubspot_owner_id",
        "hs_associated_contact_labels",
        "hs_task_is_overdue",
        "hs_task_priority",
        "hs_task_status",
        "hs_task_subject",
        "hs_task_type",
    ],
}


def get_object_properties(object_type: HubspotObject) -> list[str]:
    return HUBSPOT_OBJECT_PROPERTIES[object_type]



import logging
from typing import Any

from arcade_tdk.errors import RetryableToolError, ToolExecutionError
from google.oauth2.credentials import Credentials
from googleapiclient.discovery import Resource, build

from arcade_google_sheets.constants import (
    DEFAULT_SHEET_COLUMN_COUNT,
    DEFAULT_SHEET_ROW_COUNT,
)
from arcade_google_sheets.enums import NumberFormatType
from arcade_google_sheets.models import (
    CellData,
    CellExtendedValue,
    CellFormat,
    GridData,
    GridProperties,
    NumberFormat,
    RowData,
    Sheet,
    SheetDataInput,
    SheetProperties,
)
from arcade_google_sheets.types import CellValue

logging.basicConfig(
    level=logging.DEBUG,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)

logger = logging.getLogger(__name__)


def build_sheets_service(auth_token: str | None) -> Resource:  # type: ignore[no-any-unimported]
    """
    Build a Sheets service object.
    """
    auth_token = auth_token or ""
    return build("sheets", "v4", credentials=Credentials(auth_token))


def col_to_index(col: str) -> int:
    """Convert a sheet's column string to a 0-indexed column index

    Args:
        col (str): The column string to convert. e.g., "A", "AZ", "QED"

    Returns:
        int: The 0-indexed column index.
    """
    result = 0
    for char in col.upper():
        result = result * 26 + (ord(char) - ord("A") + 1)
    return result - 1


def index_to_col(index: int) -> str:
    """Convert a 0-indexed column index to its corresponding column string

    Args:
        index (int): The 0-indexed column index to convert.

    Returns:
        str: The column string. e.g., "A", "AZ", "QED"
    """
    result = ""
    index += 1
    while index > 0:
        index, rem = divmod(index - 1, 26)
        result = chr(rem + ord("A")) + result
    return result


def is_col_greater(col1: str, col2: str) -> bool:
    """Determine if col1 represents a column that comes after col2 in a sheet

    This comparison is based on:
      1. The length of the column string (longer means greater).
      2. Lexicographical comparison if both strings are the same length.

    Args:
        col1 (str): The first column string to compare.
        col2 (str): The second column string to compare.

    Returns:
        bool: True if col1 comes after col2, False otherwise.
    """
    if len(col1) != len(col2):
        return len(col1) > len(col2)
    return col1.upper() > col2.upper()


def compute_sheet_data_dimensions(
    sheet_data_input: SheetDataInput,
) -> tuple[tuple[int, int], tuple[int, int]]:
    """
    Compute the dimensions of a sheet based on the data provided.

    Args:
        sheet_data_input (SheetDataInput):
            The data to compute the dimensions of.

    Returns:
        tuple[tuple[int, int], tuple[int, int]]: The dimensions of the sheet. The first tuple
            contains the row range (start, end) and the second tuple contains the column range
            (start, end).
    """
    max_row = 0
    min_row = 10_000_000  # max number of cells in a sheet
    max_col_str = None
    min_col_str = None

    for key, row in sheet_data_input.data.items():
        try:
            row_num = int(key)
        except ValueError:
            continue
        if row_num > max_row:
            max_row = row_num
        if row_num < min_row:
            min_row = row_num

        if isinstance(row, dict):
            for col in row:
                # Update max column string
                if max_col_str is None or is_col_greater(col, max_col_str):
                    max_col_str = col
                # Update min column string
                if min_col_str is None or is_col_greater(min_col_str, col):
                    min_col_str = col

    max_col_index = col_to_index(max_col_str) if max_col_str is not None else -1
    min_col_index = col_to_index(min_col_str) if min_col_str is not None else 0

    return (min_row, max_row), (min_col_index, max_col_index)


def create_sheet(sheet_data_input: SheetDataInput) -> Sheet:
    """Create a Google Sheet from a dictionary of data.

    Args:
        sheet_data_input (SheetDataInput): The data to create the sheet from.

    Returns:
        Sheet: The created sheet.
    """
    (_, max_row), (min_col_index, max_col_index) = compute_sheet_data_dimensions(sheet_data_input)
    sheet_data = create_sheet_data(sheet_data_input, min_col_index, max_col_index)
    sheet_properties = create_sheet_properties(
        row_count=max(DEFAULT_SHEET_ROW_COUNT, max_row),
        column_count=max(DEFAULT_SHEET_COLUMN_COUNT, max_col_index + 1),
    )

    return Sheet(properties=sheet_properties, data=sheet_data)


def create_sheet_properties(
    sheet_id: int = 1,
    title: str = "Sheet1",
    row_count: int = DEFAULT_SHEET_ROW_COUNT,
    column_count: int = DEFAULT_SHEET_COLUMN_COUNT,
) -> SheetProperties:
    """Create a SheetProperties object

    Args:
        sheet_id (int): The ID of the sheet.
        title (str): The title of the sheet.
        row_count (int): The number of rows in the sheet.
        column_count (int): The number of columns in the sheet.

    Returns:
        SheetProperties: The created sheet properties object.
    """
    return SheetProperties(
        sheetId=sheet_id,
        title=title,
        gridProperties=GridProperties(rowCount=row_count, columnCount=column_count),
    )


def group_contiguous_rows(row_numbers: list[int]) -> list[list[int]]:
    """Groups a sorted list of row numbers into contiguous groups

    A contiguous group is a list of row numbers that are consecutive integers.
    For example, [1,2,3,5,6] is converted to [[1,2,3],[5,6]].

    Args:
        row_numbers (list[int]): The list of row numbers to group.

    Returns:
        list[list[int]]: The grouped row numbers.
    """
    if not row_numbers:
        return []
    groups = []
    current_group = [row_numbers[0]]
    for r in row_numbers[1:]:
        if r == current_group[-1] + 1:
            current_group.append(r)
        else:
            groups.append(current_group)
            current_group = [r]
    groups.append(current_group)
    return groups


def create_cell_data(cell_value: CellValue) -> CellData:
    """
    Create a CellData object based on the type of cell_value.
    """
    if isinstance(cell_value, bool):
        return _create_bool_cell(cell_value)
    elif isinstance(cell_value, int):
        return _create_int_cell(cell_value)
    elif isinstance(cell_value, float):
        return _create_float_cell(cell_value)
    elif isinstance(cell_value, str):
        return _create_string_cell(cell_value)


def _create_formula_cell(cell_value: str) -> CellData:
    cell_val = CellExtendedValue(formulaValue=cell_value)
    return CellData(userEnteredValue=cell_val)


def _create_currency_cell(cell_value: str) -> CellData:
    value_without_symbol = cell_value[1:]
    try:
        num_value = int(value_without_symbol)
        cell_format = CellFormat(
            numberFormat=NumberFormat(type=NumberFormatType.CURRENCY, pattern="$#,##0")
        )
        cell_val = CellExtendedValue(numberValue=num_value)
        return CellData(userEnteredValue=cell_val, userEnteredFormat=cell_format)
    except ValueError:
        try:
            num_value = float(value_without_symbol)  # type: ignore[assignment]
            cell_format = CellFormat(
                numberFormat=NumberFormat(type=NumberFormatType.CURRENCY, pattern="$#,##0.00")
            )
            cell_val = CellExtendedValue(numberValue=num_value)
            return CellData(userEnteredValue=cell_val, userEnteredFormat=cell_format)
        except ValueError:
            return CellData(userEnteredValue=CellExtendedValue(stringValue=cell_value))


def _create_percent_cell(cell_value: str) -> CellData:
    try:
        num_value = float(cell_value[:-1].strip())
        cell_format = CellFormat(
            numberFormat=NumberFormat(type=NumberFormatType.PERCENT, pattern="0.00%")
        )
        cell_val = CellExtendedValue(numberValue=num_value)
        return CellData(userEnteredValue=cell_val, userEnteredFormat=cell_format)
    except ValueError:
        return CellData(userEnteredValue=CellExtendedValue(stringValue=cell_value))


def _create_bool_cell(cell_value: bool) -> CellData:
    return CellData(userEnteredValue=CellExtendedValue(boolValue=cell_value))


def _create_int_cell(cell_value: int) -> CellData:
    cell_format = CellFormat(
        numberFormat=NumberFormat(type=NumberFormatType.NUMBER, pattern="#,##0")
    )
    return CellData(
        userEnteredValue=CellExtendedValue(numberValue=cell_value), userEnteredFormat=cell_format
    )


def _create_float_cell(cell_value: float) -> CellData:
    cell_format = CellFormat(
        numberFormat=NumberFormat(type=NumberFormatType.NUMBER, pattern="#,##0.00")
    )
    return CellData(
        userEnteredValue=CellExtendedValue(numberValue=cell_value), userEnteredFormat=cell_format
    )


def _create_string_cell(cell_value: str) -> CellData:
    if cell_value.startswith("="):
        return _create_formula_cell(cell_value)
    elif cell_value.startswith("$") and len(cell_value) > 1:
        return _create_currency_cell(cell_value)
    elif cell_value.endswith("%") and len(cell_value) > 1:
        return _create_percent_cell(cell_value)

    return CellData(userEnteredValue=CellExtendedValue(stringValue=cell_value))


def create_row_data(
    row_data: dict[str, CellValue], min_col_index: int, max_col_index: int
) -> RowData:
    """Constructs RowData for a single row using the provided row_data.

    Args:
        row_data (dict[str, CellValue]): The data to create the row from.
        min_col_index (int): The minimum column index from the SheetDataInput.
        max_col_index (int): The maximum column index from the SheetDataInput.
    """
    row_cells = []
    for col_idx in range(min_col_index, max_col_index + 1):
        col_letter = index_to_col(col_idx)
        if col_letter in row_data:
            cell_data = create_cell_data(row_data[col_letter])
        else:
            cell_data = CellData(userEnteredValue=CellExtendedValue(stringValue=""))
        row_cells.append(cell_data)
    return RowData(values=row_cells)


def create_sheet_data(
    sheet_data_input: SheetDataInput,
    min_col_index: int,
    max_col_index: int,
) -> list[GridData]:
    """Create grid data from SheetDataInput by grouping contiguous rows and processing cells.

    Args:
        sheet_data_input (SheetDataInput): The data to create the sheet from.
        min_col_index (int): The minimum column index from the SheetDataInput.
        max_col_index (int): The maximum column index from the SheetDataInput.

    Returns:
        list[GridData]: The created grid data.
    """
    row_numbers = list(sheet_data_input.data.keys())
    if not row_numbers:
        return []

    sorted_rows = sorted(row_numbers)
    groups = group_contiguous_rows(sorted_rows)

    sheet_data = []
    for group in groups:
        rows_data = []
        for r in group:
            current_row_data = sheet_data_input.data.get(r, {})
            row = create_row_data(current_row_data, min_col_index, max_col_index)
            rows_data.append(row)
        grid_data = GridData(
            startRow=group[0] - 1,  # convert to 0-indexed
            startColumn=min_col_index,
            rowData=rows_data,
        )
        sheet_data.append(grid_data)

    return sheet_data


def parse_get_spreadsheet_response(api_response: dict) -> dict:
    """
    Parse the get spreadsheet Google Sheets API response into a structured dictionary.
    """
    properties = api_response.get("properties", {})
    sheets = [parse_sheet(sheet) for sheet in api_response.get("sheets", [])]

    return {
        "title": properties.get("title", ""),
        "spreadsheetId": api_response.get("spreadsheetId", ""),
        "spreadsheetUrl": api_response.get("spreadsheetUrl", ""),
        "sheets": sheets,
    }


def parse_sheet(api_sheet: dict) -> dict:
    """
    Parse an individual sheet's data from the Google Sheets 'get spreadsheet'
    API response into a structured dictionary.
    """
    props = api_sheet.get("properties", {})
    grid_props = props.get("gridProperties", {})
    cell_data = convert_api_grid_data_to_dict(api_sheet.get("data", []))

    return {
        "sheetId": props.get("sheetId"),
        "title": props.get("title", ""),
        "rowCount": grid_props.get("rowCount", 0),
        "columnCount": grid_props.get("columnCount", 0),
        "data": cell_data,
    }


def extract_user_entered_cell_value(cell: dict) -> Any:
    """
    Extract the user entered value from a cell's 'userEnteredValue'.

    Args:
        cell (dict): A cell dictionary from the grid data.

    Returns:
        The extracted value if present, otherwise None.
    """
    user_val = cell.get("userEnteredValue", {})
    for key in ["stringValue", "numberValue", "boolValue", "formulaValue"]:
        if key in user_val:
            return user_val[key]

    return ""


def process_row(row: dict, start_column_index: int) -> dict:
    """
    Process a single row from grid data, converting non-empty cells into a dictionary
    that maps column letters to cell values.

    Args:
        row (dict): A row from the grid data.
        start_column_index (int): The starting column index for this row.

    Returns:
        dict: A mapping of column letters to cell values for non-empty cells.
    """
    row_result = {}
    for j, cell in enumerate(row.get("values", [])):
        column_index = start_column_index + j
        column_string = index_to_col(column_index)
        user_entered_cell_value = extract_user_entered_cell_value(cell)
        formatted_cell_value = cell.get("formattedValue", "")

        if user_entered_cell_value != "" or formatted_cell_value != "":
            row_result[column_string] = {
                "userEnteredValue": user_entered_cell_value,
                "formattedValue": formatted_cell_value,
            }

    return row_result


def convert_api_grid_data_to_dict(grids: list[dict]) -> dict:
    """
    Convert a list of grid data dictionaries from the 'get spreadsheet' API
    response into a structured cell dictionary.

    The returned dictionary maps row numbers to sub-dictionaries that map column letters
    (e.g., 'A', 'B', etc.) to their corresponding non-empty cell values.

    Args:
        grids (list[dict]): The list of grid data dictionaries from the API.

    Returns:
        dict: A dictionary mapping row numbers to dictionaries of column letter/value pairs.
            Only includes non-empty rows and non-empty cells.
    """
    result = {}
    for grid in grids:
        start_row = grid.get("startRow", 0)
        start_column = grid.get("startColumn", 0)

        for i, row in enumerate(grid.get("rowData", []), start=1):
            current_row = start_row + i
            row_data = process_row(row, start_column)

            if row_data:
                result[current_row] = row_data

    return dict(sorted(result.items()))


def validate_write_to_cell_params(  # type: ignore[no-any-unimported]
    service: Resource,
    spreadsheet_id: str,
    sheet_name: str,
    column: str,
    row: int,
) -> None:
    """Validates the input parameters for the write to cell tool.

    Args:
        service (Resource): The Google Sheets service.
        spreadsheet_id (str): The ID of the spreadsheet provided to the tool.
        sheet_name (str): The name of the sheet provided to the tool.
        column (str): The column to write to provided to the tool.
        row (int): The row to write to provided to the tool.

    Raises:
        RetryableToolError:
            If the sheet name is not found in the spreadsheet
        ToolExecutionError:
            If the column is not alphabetical
            If the row is not a positive number
            If the row is out of bounds for the sheet
            If the column is out of bounds for the sheet
    """
    if not column.isalpha():
        raise ToolExecutionError(
            message=(
                f"Invalid column name {column}. "
                "It must be a non-empty string containing only letters"
            ),
        )

    if row < 1:
        raise ToolExecutionError(
            message=(f"Invalid row number {row}. It must be a positive integer greater than 0."),
        )

    sheet_properties = (
        service.spreadsheets()
        .get(
            spreadsheetId=spreadsheet_id,
            includeGridData=True,
            fields="sheets/properties/title,sheets/properties/gridProperties/rowCount,sheets/properties/gridProperties/columnCount",
        )
        .execute()
    )
    sheet_names = [sheet["properties"]["title"] for sheet in sheet_properties["sheets"]]
    sheet_row_count = sheet_properties["sheets"][0]["properties"]["gridProperties"]["rowCount"]
    sheet_column_count = sheet_properties["sheets"][0]["properties"]["gridProperties"][
        "columnCount"
    ]

    if sheet_name not in sheet_names:
        raise RetryableToolError(
            message=f"Sheet name {sheet_name} not found in spreadsheet with id {spreadsheet_id}",
            additional_prompt_content=f"Sheet names in the spreadsheet: {sheet_names}",
            retry_after_ms=100,
        )

    if row > sheet_row_count:
        raise ToolExecutionError(
            message=(
                f"Row {row} is out of bounds for sheet {sheet_name} "
                f"in spreadsheet with id {spreadsheet_id}. "
                f"Sheet only has {sheet_row_count} rows which is less than the requested row {row}"
            )
        )

    if col_to_index(column) > sheet_column_count:
        raise ToolExecutionError(
            message=(
                f"Column {column} is out of bounds for sheet {sheet_name} "
                f"in spreadsheet with id {spreadsheet_id}. "
                f"Sheet only has {sheet_column_count} columns which "
                f"is less than the requested column {column}"
            )
        )


def parse_write_to_cell_response(response: dict) -> dict:
    return {
        "spreadsheetId": response["spreadsheetId"],
        "sheetTitle": response["updatedData"]["range"].split("!")[0],
        "updatedCell": response["updatedData"]["range"].split("!")[1],
        "value": response["updatedData"]["values"][0][0],
    }



from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    SimilarityCritic,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_google_docs
from arcade_google_docs.enum import DocumentFormat, OrderBy
from arcade_google_docs.tools import (
    create_blank_document,
    create_document_from_text,
    get_document_by_id,
    insert_text_at_end_of_document,
    search_and_retrieve_documents,
    search_documents,
)

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.9,
    warn_threshold=0.95,
)

catalog = ToolCatalog()
catalog.add_module(arcade_google_docs)


@tool_eval()
def docs_eval_suite() -> EvalSuite:
    """Create an evaluation suite for Google Docs tools."""
    suite = EvalSuite(
        name="Google Docs Tools Evaluation",
        system_message="You are an AI assistant that can create and manage Google Docs using the provided tools.",
        catalog=catalog,
        rubric=rubric,
    )

    # A previous tool call to list_documents
    additional_messages = [
        {"role": "user", "content": "list my 10 most recently created docs"},
        {
            "role": "assistant",
            "content": "Please go to this URL and authorize the action: [Link](https://accounts.google.com/)",
        },
        {
            "role": "assistant",
            "content": "",
            "tool_calls": [
                {
                    "id": "call_gegK723W2hXsORjBmq1Oexqk",
                    "type": "function",
                    "function": {
                        "name": "Google_ListDocuments",
                        "arguments": '{"limit":10,"order_by":"createdTime desc"}',
                    },
                }
            ],
        },
        {
            "role": "tool",
            "content": '{"documents":[{"id":"1e0rCoT1Yd14WuuEvd3hSUcN_-VD3df4T3Q08uLm3TWc","kind":"drive#file","mimeType":"application/vnd.google-apps.document","name":"Tst10"},{"id":"1eTSWd-5zQds8K9OWYygwtCFMUyuuMize3bh3HaRsKts","kind":"drive#file","mimeType":"application/vnd.google-apps.document","name":"Tst9"},{"id":"19Dqugn0rVi89K0C__lpg1HbhQOTenccyZOhPgivTHMs","kind":"drive#file","mimeType":"application/vnd.google-apps.document","name":"Tst8"},{"id":"1RCibzx14eqP3vS9yI4nD13OKf8Vee56RiszS53OkR7I","kind":"drive#file","mimeType":"application/vnd.google-apps.document","name":"Tst7"},{"id":"1imFb04JQuBn8SiSsRFf6fEuYCyXkbII4KX8fsmnT0jo","kind":"drive#file","mimeType":"application/vnd.google-apps.document","name":"Tst6"},{"id":"1ZC3oypdfLWFgBd-emeSykJf9tZOae6USsFboygRCr-w","kind":"drive#file","mimeType":"application/vnd.google-apps.document","name":"Tst5"},{"id":"1-gFGNWmwLxEiKa6NNixLNq3X-phXRMORVZfVTfBg8Sc","kind":"drive#file","mimeType":"application/vnd.google-apps.document","name":"Tst4"},{"id":"1eQ8UBO_PY3Lem4R8OVdIc9ODXt0MrSUAnEu994Qz8P8","kind":"drive#file","mimeType":"application/vnd.google-apps.document","name":"Tst3"},{"id":"1TOxB0MLry-JzntDWDT1LFywTLdr3XDWPT5L5UsHMs5c","kind":"drive#file","mimeType":"application/vnd.google-apps.document","name":"Tst2"},{"id":"1a1UQ7C90s8kGfnO8k6wfAZz_Cy5nGN2MkCoRB5y2j3w","kind":"drive#file","mimeType":"application/vnd.google-apps.document","name":"Tst1"}],"documents_count":10}',
            "tool_call_id": "call_gegK723W2hXsORjBmq1Oexqk",
            "name": "Google_ListDocuments",
        },
        {
            "role": "assistant",
            "content": "Here are your 10 most recently created Google Docs:\n\n1. [Tst10](https://docs.google.com/document/d/1e0rCoT1Yd14WuuEvd3hSUcN_-VD3df4T3Q08uLm3TWc)\n2. [Tst9](https://docs.google.com/document/d/1eTSWd-5zQds8K9OWYygwtCFMUyuuMize3bh3HaRsKts)\n3. [Tst8](https://docs.google.com/document/d/19Dqugn0rVi89K0C__lpg1HbhQOTenccyZOhPgivTHMs)\n4. [Tst7](https://docs.google.com/document/d/1RCibzx14eqP3vS9yI4nD13OKf8Vee56RiszS53OkR7I)\n5. [Tst6](https://docs.google.com/document/d/1imFb04JQuBn8SiSsRFf6fEuYCyXkbII4KX8fsmnT0jo)\n6. [Tst5](https://docs.google.com/document/d/1ZC3oypdfLWFgBd-emeSykJf9tZOae6USsFboygRCr-w)\n7. [Tst4](https://docs.google.com/document/d/1-gFGNWmwLxEiKa6NNixLNq3X-phXRMORVZfVTfBg8Sc)\n8. [Tst3](https://docs.google.com/document/d/1eQ8UBO_PY3Lem4R8OVdIc9ODXt0MrSUAnEu994Qz8P8)\n9. [Tst2](https://docs.google.com/document/d/1TOxB0MLry-JzntDWDT1LFywTLdr3XDWPT5L5UsHMs5c)\n10. [Tst1](https://docs.google.com/document/d/1a1UQ7C90s8kGfnO8k6wfAZz_Cy5nGN2MkCoRB5y2j3w)\n\nYou can click the links to open each document.",
        },
    ]

    suite.add_case(
        name="Get document content",
        user_message="Can you read me the contents of Tst9 doc and also Tst10 doc please",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_document_by_id,
                args={
                    "document_id": "1eTSWd-5zQds8K9OWYygwtCFMUyuuMize3bh3HaRsKts",
                },
            ),
            ExpectedToolCall(
                func=get_document_by_id,
                args={
                    "document_id": "1e0rCoT1Yd14WuuEvd3hSUcN_-VD3df4T3Q08uLm3TWc",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="document_id", weight=0.6),
        ],
        additional_messages=additional_messages,
    )

    suite.add_case(
        name="Insert text at end of document",
        user_message="Please add the text 'This is a new paragraph.' to the end of Tst4.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=insert_text_at_end_of_document,
                args={
                    "document_id": "1-gFGNWmwLxEiKa6NNixLNq3X-phXRMORVZfVTfBg8Sc",
                    "text_content": "This is a new paragraph.",
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="document_id", weight=0.5),
            SimilarityCritic(critic_field="text_content", weight=0.5),
        ],
        additional_messages=additional_messages,
    )

    suite.add_case(
        name="Read the contents of two documents and then insert text at end of a different document.",
        user_message="Can you read me the contents of Tst9 doc and also Tst10 doc please. Also, please add the text 'This is a new paragraph.' to the end of Tst4.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=insert_text_at_end_of_document,
                args={
                    "document_id": "1-gFGNWmwLxEiKa6NNixLNq3X-phXRMORVZfVTfBg8Sc",
                    "text_content": "This is a new paragraph.",
                },
            ),
            ExpectedToolCall(
                func=get_document_by_id,
                args={
                    "document_id": "1eTSWd-5zQds8K9OWYygwtCFMUyuuMize3bh3HaRsKts",
                },
            ),
            ExpectedToolCall(
                func=get_document_by_id,
                args={
                    "document_id": "1e0rCoT1Yd14WuuEvd3hSUcN_-VD3df4T3Q08uLm3TWc",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="document_id", weight=0.3),
            SimilarityCritic(critic_field="text_content", weight=0.3),
        ],
        additional_messages=additional_messages,
    )

    suite.add_case(
        name="Create blank document",
        user_message="Create a new Doc titled 'Meeting Notes'.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=create_blank_document,
                args={
                    "title": "Meeting Notes",
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="title", weight=1.0),
        ],
    )

    suite.add_case(
        name="Create document from text",
        user_message="Create a new doc called To-Do List with the content 'Buy groceries, Call mom, Finish report'.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=create_document_from_text,
                args={
                    "title": "To-Do List",
                    "text_content": "Buy groceries\nCall mom\nFinish report",
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="title", weight=0.5),
            SimilarityCritic(critic_field="text_content", weight=0.5),
        ],
    )

    suite.add_case(
        name="No tool call case",
        user_message="Create a new microsoft word document titled 'My Resume'.",
        expected_tool_calls=[],
        critics=[],
    )

    return suite


@tool_eval()
def search_documents_eval_suite() -> EvalSuite:
    """Create an evaluation suite for Google Drive tools."""
    suite = EvalSuite(
        name="Google Drive Tools Evaluation",
        system_message="You are an AI assistant that can manage Google Drive documents using the provided tools.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Search documents in Google Drive",
        user_message="get my 49 most recently created documents, list the ones created most recently first.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_documents,
                args={
                    "order_by": [OrderBy.CREATED_TIME_DESC.value],
                    "limit": 49,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="order_by", weight=0.5),
            BinaryCritic(critic_field="limit", weight=0.5),
        ],
    )

    suite.add_case(
        name="Search documents in Google Drive based on document keywords",
        user_message="Search the documents that contain the word 'greedy' and the phrase 'hello, world'",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_documents,
                args={
                    "document_contains": ["greedy", "hello, world"],
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="document_contains", weight=1.0),
        ],
    )

    suite.add_case(
        name="Search documents in a specific Google Drive based on document keywords",
        user_message="Search the documents that contain the word 'greedy' and the phrase 'hello, world' in the drive with id 'abc123'",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_documents,
                args={
                    "document_contains": ["greedy", "hello, world"],
                    "search_only_in_shared_drive_id": "abc123",
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="search_only_in_shared_drive_id", weight=0.5),
            BinaryCritic(critic_field="document_contains", weight=0.5),
        ],
    )

    suite.add_case(
        name="Search documents in a Google Drive Workspace organization domain based on document keywords",
        user_message="Search the documents that contain the phrase 'hello, world' in the organization domain",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_documents,
                args={
                    "document_contains": ["hello, world"],
                    "include_organization_domain_documents": True,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="include_organization_domain_documents", weight=0.5),
            BinaryCritic(critic_field="document_contains", weight=0.5),
        ],
    )

    suite.add_case(
        name="Search documents in shared drives",
        user_message="Search the 5 documents from all drives corpora that nobody has touched in forever, excluding shared drives.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_documents,
                args={
                    "limit": 5,
                    "include_shared_drives": False,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="include_shared_drives", weight=0.5),
            BinaryCritic(critic_field="limit", weight=0.5),
        ],
    )

    suite.add_case(
        name="No tool call case",
        user_message="List my 10 most recently modified documents that are stored in my Microsoft OneDrive.",
        expected_tool_calls=[],
        critics=[],
    )

    return suite


@tool_eval()
def search_and_retrieve_documents_eval_suite() -> EvalSuite:
    """Create an evaluation suite for Google Drive search and retrieve tools."""
    suite = EvalSuite(
        name="Google Drive Tools Evaluation",
        system_message="You are an AI assistant that can manage Google Drive documents using the provided tools.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Search and retrieve (write summary)",
        user_message="Write a summary of the documents in my Google Drive about 'MX Engineering'",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_and_retrieve_documents,
                args={
                    "document_contains": ["MX Engineering"],
                    "return_format": DocumentFormat.MARKDOWN,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="document_contains", weight=0.5),
            BinaryCritic(critic_field="return_format", weight=0.5),
        ],
    )

    suite.add_case(
        name="Search and retrieve (project proposal)",
        user_message="Display the document contents in HTML format from my Google Drive that contain the phrase 'project proposal'.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_and_retrieve_documents,
                args={
                    "document_contains": ["project proposal"],
                    "return_format": DocumentFormat.HTML,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="document_contains", weight=0.5),
            BinaryCritic(critic_field="return_format", weight=0.5),
        ],
    )

    suite.add_case(
        name="Search and retrieve (meeting notes)",
        user_message="Retrieve documents that contain both 'meeting notes' and 'budget' in JSON format.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_and_retrieve_documents,
                args={
                    "document_contains": ["meeting notes", "budget"],
                    "return_format": DocumentFormat.GOOGLE_API_JSON,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="document_contains", weight=0.5),
            BinaryCritic(critic_field="return_format", weight=0.5),
        ],
    )

    suite.add_case(
        name="Search and retrieve (Q1 report)",
        user_message="Show me the content of the documents that mention 'Q1 report' but do not include the expression 'Project XYZ'.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_and_retrieve_documents,
                args={
                    "document_contains": ["Q1 report"],
                    "document_not_contains": ["Project XYZ"],
                    "return_format": DocumentFormat.MARKDOWN,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="document_contains", weight=1 / 3),
            BinaryCritic(critic_field="document_not_contains", weight=1 / 3),
            BinaryCritic(critic_field="return_format", weight=1 / 3),
        ],
    )

    return suite



import importlib.util
import ipaddress
import os
import shlex
import webbrowser
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from importlib import metadata
from pathlib import Path
from textwrap import dedent
from typing import Any, Callable, Union, cast
from urllib.parse import urlencode, urlparse

import idna
import typer
from arcade_core import ToolCatalog, Toolkit
from arcade_core.config_model import Config
from arcade_core.errors import ToolkitLoadError
from arcade_core.schema import ToolDefinition
from arcadepy import NOT_GIVEN, APIConnectionError, APIStatusError, APITimeoutError, Arcade
from arcadepy.types import AuthorizationResponse
from openai import OpenAI, Stream
from openai.types.chat.chat_completion import Choice as ChatCompletionChoice
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk
from openai.types.chat.chat_completion_chunk import Choice as ChatCompletionChunkChoice
from pydantic import ValidationError
from rich.console import Console
from rich.live import Live
from rich.markdown import Markdown
from rich.text import Text
from typer.core import TyperGroup
from typer.models import Context

from arcade_cli.constants import LOCAL_AUTH_CALLBACK_PORT, LOCALHOST

console = Console()


# -----------------------------------------------------------------------------
# Shared helpers for the CLI
# -----------------------------------------------------------------------------


class OrderCommands(TyperGroup):
    def list_commands(self, ctx: Context) -> list[str]:  # type: ignore[override]
        """Return list of commands in the order appear."""
        return list(self.commands)  # get commands using self.commands


class ChatCommand(str, Enum):
    HELP = "/help"
    HELP_ALT = "/?"
    CLEAR = "/clear"
    HISTORY = "/history"
    SHOW = "/show"
    EXIT = "/exit"


def create_cli_catalog(
    toolkit: str | None = None,
    show_toolkits: bool = False,
) -> ToolCatalog:
    """
    Load toolkits from the python environment.
    """
    if toolkit:
        toolkit = toolkit.lower().replace("-", "_")
        try:
            prefixed_toolkit = "arcade_" + toolkit
            toolkits = [Toolkit.from_package(prefixed_toolkit)]
        except ToolkitLoadError:
            try:  # try without prefix
                toolkits = [Toolkit.from_package(toolkit)]
            except ToolkitLoadError as e:
                console.print(f"❌ {e}", style="bold red")
                typer.Exit(code=1)
    else:
        toolkits = Toolkit.find_all_arcade_toolkits()

    if not toolkits:
        console.print("❌ No toolkits found or specified", style="bold red")
        typer.Exit(code=1)

    catalog = ToolCatalog()
    for loaded_toolkit in toolkits:
        if show_toolkits:
            console.print(f"Loading toolkit: {loaded_toolkit.name}", style="bold blue")
        catalog.add_toolkit(loaded_toolkit)
    return catalog


def compute_base_url(
    force_tls: bool,
    force_no_tls: bool,
    host: str,
    port: int | None,
) -> str:
    """
    Compute the base URL for the Arcade Engine from the provided overrides.

    Treats 127.0.0.1 and 0.0.0.0 as aliases for localhost.

    force_no_tls takes precedence over force_tls. For example, if both are set to True,
    the resulting URL will use http.

    The port is included in the URL unless the host is a fully qualified domain name
    (excluding IP addresses) and no port is specified. Handles IPv4, IPv6, IDNs, and
    hostnames with underscores.

    This property exists to provide a consistent and correctly formatted URL for
    connecting to the Arcade Engine, taking into account various configuration
    options and edge cases. It ensures that:

    1. The correct protocol (http/https) is used based on the TLS setting.
    2. IPv4 and IPv6 addresses are properly formatted.
    3. Internationalized Domain Names (IDNs) are correctly encoded.
    4. Fully Qualified Domain Names (FQDNs) are identified and handled appropriately.
    5. Ports are included when necessary, respecting common conventions for FQDNs.
    6. Hostnames with underscores (common in development environments) are supported.
    7. Pre-existing port specifications in the host are respected.

    The resulting URL is always suffixed with api_version to specify the API version.

    Returns:
        str: The fully constructed URL for the Arcade Engine.
    """
    # "Use 127.0.0.1" and "0.0.0.0" as aliases for "localhost"
    host = LOCALHOST if host in ["127.0.0.1", "0.0.0.0"] else host  # noqa: S104

    # Determine TLS setting based on input flags
    if force_no_tls:
        is_tls = False
    elif force_tls:
        is_tls = True
    else:
        is_tls = host != LOCALHOST

    # "localhost" defaults to dev port if not specified
    if host == LOCALHOST and port is None:
        port = 9099

    protocol = "https" if is_tls else "http"

    # Handle potential IDNs
    try:
        encoded_host = idna.encode(host).decode("ascii")
    except idna.IDNAError:
        encoded_host = host

    # Check if the host is a valid IP address (IPv4 or IPv6)
    try:
        ipaddress.ip_address(encoded_host)
        is_ip = True
    except ValueError:
        is_ip = False

    # Parse the host, handling potential IPv6 addresses
    host_for_parsing = f"[{encoded_host}]" if is_ip and ":" in encoded_host else encoded_host
    parsed_host = urlparse(f"//{host_for_parsing}")

    # Check if the host is a fully qualified domain name (excluding IP addresses)
    is_fqdn = "." in parsed_host.netloc and not is_ip and "_" not in parsed_host.netloc

    # Handle hosts that might already include a port
    if ":" in parsed_host.netloc and not is_ip:
        host, existing_port = parsed_host.netloc.rsplit(":", 1)
        if existing_port.isdigit():
            return f"{protocol}://{parsed_host.netloc}"

    if is_fqdn and port is None:
        return f"{protocol}://{encoded_host}"
    elif port is not None:
        return f"{protocol}://{encoded_host}:{port}"
    else:
        return f"{protocol}://{encoded_host}"


def compute_login_url(
    host: str, state: str, port: int | None, callback_host: str | None = None
) -> str:
    """
    Compute the full URL for the CLI login endpoint.
    """
    if callback_host:
        if not (callback_host.startswith("http://") or callback_host.startswith("https://")):
            callback_uri = f"http://{callback_host}"
        else:
            callback_uri = callback_host
        if not callback_uri.rstrip("/").endswith("/callback"):
            if port:
                callback_uri = callback_uri.rstrip("/") + f":{port}" + "/callback"
            else:
                callback_uri = callback_uri.rstrip("/") + "/callback"
    else:
        callback_uri = f"http://{LOCALHOST}:{LOCAL_AUTH_CALLBACK_PORT}/callback"

    params = urlencode({"callback_uri": callback_uri, "state": state})

    port = port if port else 8000

    login_base_url = (
        f"http://{LOCALHOST}:{port}"
        if host in [LOCALHOST, "127.0.0.1", "0.0.0.0"]  # noqa: S104
        else f"https://{host}"
    )
    endpoint = "/api/v1/auth/cli_login"

    return f"{login_base_url}{endpoint}?{params}"


def get_tools_from_engine(
    host: str,
    port: int | None = None,
    force_tls: bool = False,
    force_no_tls: bool = False,
    toolkit: str | None = None,
) -> list[ToolDefinition]:
    config = validate_and_get_config()
    base_url = compute_base_url(force_tls, force_no_tls, host, port)
    client = Arcade(api_key=config.api.key, base_url=base_url)

    tools = []
    try:
        page_iterator = client.tools.list(toolkit=toolkit or NOT_GIVEN)
        for tool in page_iterator:
            try:
                tools.append(ToolDefinition.model_validate(tool.model_dump()))
            except ValidationError:
                # Skip listing tools that aren't valid ToolDefinitions
                continue
    except APIConnectionError:
        console.print(
            f"❌ Can't connect to Arcade Engine at {base_url}. (Is it running?)", style="bold red"
        )

    return tools


def get_tool_messages(choice: dict) -> list[dict]:
    if hasattr(choice, "tool_messages") and choice.tool_messages:
        return choice.tool_messages  # type: ignore[no-any-return]
    return []


@dataclass
class StreamingResult:
    role: str
    full_message: str
    tool_messages: list
    tool_authorization: dict | None


def handle_streaming_content(stream: Stream[ChatCompletionChunk], model: str) -> StreamingResult:
    """
    Display the streamed markdown chunks as a single line.
    """
    from rich.live import Live

    full_message = ""
    tool_messages = []
    tool_authorization = None
    role = ""
    printed_role: bool = False

    with Live(console=console, refresh_per_second=10) as live:
        for chunk in stream:
            choice = chunk.choices[0]
            role = choice.delta.role or role

            # Display and get tool messages if they exist
            tool_messages += get_tool_messages(choice)  # type: ignore[arg-type]
            tool_authorization = get_tool_authorization(choice)

            chunk_message = choice.delta.content

            if role == "assistant" and tool_authorization:
                continue  # Skip the message if it's an auth request (handled later in handle_tool_authorization)

            if role == "assistant" and not printed_role:
                console.print(f"\n[blue][bold]Assistant[/bold] ({model}):[/blue] ")
                printed_role = True

            if chunk_message:
                full_message += chunk_message
                markdown_chunk = Markdown(full_message)
                live.update(markdown_chunk)

        # Markdownify URLs in the final message if applicable
        if role == "assistant":
            full_message = markdownify_urls(full_message)
            live.update(Markdown(full_message))

    return StreamingResult(role, full_message, tool_messages, tool_authorization)


def markdownify_urls(message: str) -> str:
    """
    Convert URLs in the message to markdown links.
    """
    import re

    # This regex will match URLs that are not already formatted as markdown links:
    # [Link text](https://example.com)
    url_pattern = r"(?<!\]\()https?://\S+"

    # Wrap all URLs in the message with markdown links
    return re.sub(url_pattern, r"[Link](\g<0>)", message)


def validate_and_get_config(
    validate_api: bool = True,
    validate_user: bool = True,
) -> Config:
    """
    Validates the configuration, user, and returns the Config object
    """
    from arcade_core.config import config

    if validate_api and (not config.api or not config.api.key):
        console.print(
            "❌ API configuration not found or key is missing. Please run `arcade login`.",
            style="bold red",
        )
        raise typer.Exit(code=1)

    if validate_user and (not config.user or not config.user.email):
        console.print(
            "❌ User email not found in configuration. Please run `arcade login`.", style="bold red"
        )
        raise typer.Exit(code=1)

    return config


def log_engine_health(client: Arcade) -> None:
    try:
        result = client.health.check(timeout=2)
        if result.healthy:
            return

        console.print(
            "⚠️ Warning: Arcade Engine is unhealthy",
            style="bold yellow",
        )

    except APIConnectionError:
        console.print(
            "⚠️ Warning: Arcade Engine was unreachable. (Is it running?)",
            style="bold yellow",
        )

    except APIStatusError as e:
        console.print(
            "[bold][yellow]⚠️ Warning: "
            + str(e)
            + " ("
            + "[/yellow]"
            + "[red]"
            + str(e.status_code)
            + "[/red]"
            + "[yellow])[/yellow][/bold]"
        )


@dataclass
class ChatInteractionResult:
    history: list[dict]
    tool_messages: list[dict]
    tool_authorization: dict | None


def handle_chat_interaction(
    client: OpenAI, model: str, history: list[dict], user_email: str | None, stream: bool = False
) -> ChatInteractionResult:
    """
    Handle a single chat-request/chat-response interaction for both streamed and non-streamed responses.
    Handling the chat response includes:
    - Streaming the response if the stream flag is set
    - Displaying the response in the console
    - Getting the tool messages and tool authorization from the response
    - Updating the history with the response, tool calls, and tool responses
    """
    if stream:
        # TODO Fix this in the client so users don't deal with these
        # typing issues
        response = client.chat.completions.create(  # type: ignore[call-overload]
            model=model,
            messages=history,
            tool_choice="generate",
            user=user_email,
            stream=True,
        )
        streaming_result = handle_streaming_content(response, model)
        role, message_content = streaming_result.role, streaming_result.full_message
        tool_messages, tool_authorization = (
            streaming_result.tool_messages,
            streaming_result.tool_authorization,
        )
    else:
        response = client.chat.completions.create(  # type: ignore[call-overload]
            model=model,
            messages=history,
            tool_choice="generate",
            user=user_email,
            stream=False,
        )
        message_content = response.choices[0].message.content or ""

        # Get extra fields from the response
        tool_messages = get_tool_messages(response.choices[0])
        tool_authorization = get_tool_authorization(response.choices[0])

        role = response.choices[0].message.role

        if role == "assistant" and tool_authorization:
            pass  # Skip the message if it's an auth request (handled later in handle_tool_authorization)
        elif role == "assistant":
            message_content = markdownify_urls(message_content)
            console.print(
                f"\n[blue][bold]Assistant[/bold] ({model}):[/blue] ", Markdown(message_content)
            )
        else:
            console.print(f"\n[bold]{role}:[/bold] {message_content}")

    history += tool_messages
    history.append({"role": role, "content": message_content})

    return ChatInteractionResult(history, tool_messages, tool_authorization)


def handle_tool_authorization(
    arcade_client: Arcade,
    tool_authorization: AuthorizationResponse,
    history: list[dict[str, Any]],
    openai_client: OpenAI,
    model: str,
    user_email: str | None,
    stream: bool,
) -> ChatInteractionResult:
    with Live(console=console, refresh_per_second=4) as live:
        if tool_authorization.url:
            authorization_url = str(tool_authorization.url)
            webbrowser.open(authorization_url)
            message = (
                "You'll need to authorize this action in your browser.\n\n"
                f"If a browser doesn't open automatically, click [this link]({authorization_url}) "
                f"or copy this URL and paste it into your browser:\n\n{authorization_url}"
            )
            live.update(Markdown(message, style="dim"))

        wait_for_authorization_completion(arcade_client, tool_authorization)

        message = "Thanks for authorizing the action! Sending your request..."
        live.update(Text(message, style="dim"))

    history.pop()
    return handle_chat_interaction(openai_client, model, history, user_email, stream)


def wait_for_authorization_completion(
    client: Arcade, tool_authorization: AuthorizationResponse | None
) -> None:
    """
    Wait for the authorization for a tool call to complete i.e., wait for the user to click on
    the approval link and authorize Arcade.
    """
    if tool_authorization is None:
        return

    auth_response = AuthorizationResponse.model_validate(tool_authorization)

    while auth_response.status != "completed":
        try:
            auth_response = client.auth.status(
                id=cast(str, auth_response.id),
                wait=59,
            )
        except APITimeoutError:
            continue


def get_tool_authorization(
    choice: Union[ChatCompletionChoice, ChatCompletionChunkChoice],
) -> dict | None:
    """
    Get the tool authorization from a chat response's choice.
    """
    if hasattr(choice, "tool_authorizations") and choice.tool_authorizations:
        return choice.tool_authorizations[0]  # type: ignore[no-any-return]
    return None


def is_authorization_pending(tool_authorization: dict | None) -> bool:
    """
    Check if the authorization for a tool call is pending.
    Expects a chat response's choice.tool_authorizations as input.
    """
    is_auth_pending = (
        tool_authorization is not None and tool_authorization.get("status", "") == "pending"
    )
    return is_auth_pending


def get_eval_files(directory: str) -> list[Path]:
    """
    Get a list of evaluation files starting with 'eval_' and ending with '.py' in the given directory.

    Args:
        directory: The directory to search for evaluation files.

    Returns:
        A list of Paths to the evaluation files. Returns an empty list if no files are found.
    """
    directory_path = Path(directory).resolve()

    if directory_path.is_dir():
        eval_files = [f for f in directory_path.rglob("eval_*.py") if f.is_file()]
    elif directory_path.is_file():
        eval_files = (
            [directory_path]
            if directory_path.name.startswith("eval_") and directory_path.name.endswith(".py")
            else []
        )
    else:
        console.print(f"Path not found: {directory_path}", style="bold red")
        return []

    if not eval_files:
        console.print(
            "No evaluation files found. Filenames must start with 'eval_' and end with '.py'.",
            style="bold yellow",
        )
        return []

    return eval_files


def load_eval_suites(eval_files: list[Path]) -> list[Callable]:
    """
    Load evaluation suites from the given eval_files by importing the modules
    and extracting functions decorated with `@tool_eval`.

    Args:
        eval_files: A list of Paths to evaluation files.

    Returns:
        A list of callable evaluation suite functions.
    """
    eval_suites = []
    for eval_file_path in eval_files:
        module_name = eval_file_path.stem  # filename without extension

        # Now we need to load the module from eval_file_path
        file_path_str = str(eval_file_path)
        module_name_str = module_name

        # Load using importlib
        spec = importlib.util.spec_from_file_location(module_name_str, file_path_str)
        if spec is None:
            console.print(f"Failed to load {eval_file_path}", style="bold red")
            continue

        module = importlib.util.module_from_spec(spec)
        if spec.loader is not None:
            spec.loader.exec_module(module)
        else:
            console.print(f"Failed to load module: {module_name}", style="bold red")
            continue

        eval_suite_funcs = [
            obj
            for name, obj in module.__dict__.items()
            if callable(obj) and hasattr(obj, "__tool_eval__")
        ]

        if not eval_suite_funcs:
            console.print(f"No @tool_eval functions found in {eval_file_path}", style="bold yellow")
            continue

        eval_suites.extend(eval_suite_funcs)

    return eval_suites


def get_user_input() -> str:
    """
    Get input from the user, handling multi-line input.
    """
    MULTI_LINE_PROMPT = '"""'
    user_input = input()
    # Handle multi-line input
    if user_input.startswith(MULTI_LINE_PROMPT):
        user_input = user_input[len(MULTI_LINE_PROMPT) :]

        while not user_input.endswith(MULTI_LINE_PROMPT):
            line = input()
            if not line:
                print()
            user_input += "\n" + line

        user_input = user_input.rstrip(MULTI_LINE_PROMPT)
    else:
        # Handle single-line input
        while not user_input.strip():
            user_input = input()

    return user_input.strip()


def display_chat_help() -> None:
    """Display the help message for arcade chat."""
    help_message = dedent(f"""\
        [default]
        Available Commands:
          {ChatCommand.SHOW.value:<13} Show all available tools
          {ChatCommand.HISTORY.value:<13} Show the chat history
          {ChatCommand.CLEAR.value:<13} Clear the chat history
          {ChatCommand.EXIT.value:<13} Exit the chat
          {ChatCommand.HELP_ALT.value}, {ChatCommand.HELP.value:<9} Help for a command

        Surround in \"\"\" for multi-line messages[/default]
    """)
    console.print(help_message)


def handle_user_command(
    user_input: str,
    history: list,
    host: str,
    port: int,
    force_tls: bool,
    force_no_tls: bool,
    show: Callable,
) -> bool:
    """
    Handle user commands during `arcade chat` and return True if a command was processed, otherwise False.
    """
    if user_input in [ChatCommand.HELP, ChatCommand.HELP_ALT]:
        display_chat_help()
        return True
    elif user_input == ChatCommand.EXIT:
        raise KeyboardInterrupt
    elif user_input == ChatCommand.HISTORY:
        console.print(history)
        return True
    elif user_input == ChatCommand.CLEAR:
        console.print("Chat history cleared.", style="bold green")
        history.clear()
        return True
    elif user_input == ChatCommand.SHOW:
        show(
            toolkit=None,
            tool=None,
            host=host,
            local=False,
            port=port,
            force_tls=force_tls,
            force_no_tls=force_no_tls,
            debug=False,
        )
        return True
    return False


def parse_user_command(user_input: str) -> ChatCommand | None:
    """
    Parse the user command and return the corresponding ChatCommand enum.
    Returns None if the input is not a valid chat command.
    """
    try:
        return ChatCommand(user_input)
    except ValueError:
        return None


def version_callback(value: bool) -> None:
    """Callback implementation for the `arcade --version`.
    Prints the version of Arcade and exit.
    """
    if value:
        version = metadata.version("arcade-ai")
        console.print(f"[bold]Arcade CLI[/bold] (version {version})")
        exit()


def get_today_context() -> str:
    today = datetime.now().strftime("%Y-%m-%d")
    day_of_week = datetime.now().strftime("%A")
    return f"Today is {today}, {day_of_week}."


def discover_toolkits() -> list[Toolkit]:
    """Return all Arcade toolkits installed in the active Python environment.

    Raises:
        RuntimeError: If no toolkits are found, mirroring the behaviour of Toolkit discovery elsewhere.
    """
    toolkits = Toolkit.find_all_arcade_toolkits()
    if not toolkits:
        raise RuntimeError("No toolkits found in Python environment.")
    return toolkits


def build_tool_catalog(toolkits: list[Toolkit]) -> ToolCatalog:
    """Construct a ``ToolCatalog`` populated with *toolkits*.


    Args:
        toolkits: Toolkits to register in the catalog.

    Returns:
        ToolCatalog
    """
    catalog = ToolCatalog()
    for tk in toolkits:
        catalog.add_toolkit(tk)
    return catalog


def _parse_line(line: str) -> tuple[str, str] | None:
    """
    Return (key, value) if the line looks like KEY=VALUE, else None.
    Handles quotes and escaped chars via shlex.
    """
    if not line or line.startswith("#") or "=" not in line:
        return None
    key, raw_val = line.split("=", 1)
    key = key.strip()
    raw_val = raw_val.strip()

    # Use shlex to handle "quoted strings with # hash" etc.
    try:
        value = shlex.split(raw_val)[0] if raw_val else ""
    except ValueError:
        # Fallback: naked value without shlex parsing
        value = raw_val

    return key, value


def load_dotenv(path: str | Path, *, override: bool = False) -> dict[str, str]:
    """
    Load variables from *path* into os.environ.

    Args:
        path: .env file path
        override: replace existing env vars if True

    Returns:
        The mapping of vars that were added/updated.
    """
    path = Path(path).expanduser()
    if not path.is_file():
        return {}

    loaded: dict[str, str] = {}

    for raw in path.read_text().splitlines():
        parsed = _parse_line(raw.strip())
        if parsed is None:
            continue
        k, v = parsed
        if override or k not in os.environ:
            os.environ[k] = v
            loaded[k] = v

    return loaded


def require_dependency(
    package_name: str,
    command_name: str,
    install_command: str,
) -> None:
    """
    Display a helpful error message if the required dependency is missing.

    Args:
        package_name: The name of the package to import (e.g., 'arcade_serve')
        command_name: The command that requires the package (e.g., 'serve')
        install_command: The command to install the package (e.g., "pip install 'arcade-ai[evals]'")
    """
    try:
        importlib.import_module(package_name.replace("-", "_"))
    except ImportError:
        console.print(
            f"❌ The '{package_name}' package is required to run the '{command_name}' command but is not installed.",
            style="bold red",
        )
        console.print(
            f"To install it, run the following command:\n* [green]{install_command}[/green]",
            style="bold",
        )
        raise typer.Exit(code=1)



NOTION_API_URL = "https://api.notion.com/v1"


ENDPOINTS = {
    "create_a_page": "/pages",
    "retrieve_block_children": "/blocks/{block_id}/children",
    "search_by_title": "/search",
    "query_a_database": "/databases/{database_id}/query",
    "update_page_properties": "/pages/{page_id}",
    "append_block_children": "/blocks/{block_id}/children",
    "retrieve_a_database": "/databases/{database_id}",
    "create_comment": "/comments",
    "retrieve_a_page": "/pages/{page_id}",
    "retrieve_a_block": "/blocks/{block_id}",
}

UNTITLED_TITLE = "New Page"



import json
from datetime import timedelta

from arcade_evals import (
    BinaryCritic,
    DatetimeCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_slack
from arcade_slack.critics import RelativeTimeBinaryCritic
from arcade_slack.tools.chat import get_messages

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.8,
    warn_threshold=0.9,
)


catalog = ToolCatalog()
# Register the Slack tools
catalog.add_module(arcade_slack)


@tool_eval()
def get_messages_in_channel_eval_suite() -> EvalSuite:
    """Create an evaluation suite for tools getting messages in channels."""
    suite = EvalSuite(
        name="Slack Chat Tools Evaluation",
        system_message="You are an AI assistant that can interact with Slack to send messages and get information from conversations, users, etc.",
        catalog=catalog,
        rubric=rubric,
    )

    no_arguments_user_messages_by_channel_name = [
        "what are the latest messages in the #general channel",
        "show me the messages in the general channel",
        "list the messages in the #general channel",
        "list the messages in the general channel",
    ]

    for i, user_message in enumerate(no_arguments_user_messages_by_channel_name):
        suite.add_case(
            name=f"Get messages in conversation by name {i}: '{user_message}'",
            user_message=user_message,
            expected_tool_calls=[
                ExpectedToolCall(
                    func=get_messages,
                    args={
                        "conversation_id": None,
                        "channel_name": "general",
                        "user_ids": None,
                        "usernames": None,
                        "emails": None,
                    },
                ),
            ],
            critics=[
                BinaryCritic(critic_field="conversation_id", weight=0.1),
                BinaryCritic(critic_field="channel_name", weight=0.6),
                BinaryCritic(critic_field="user_ids", weight=0.1),
                BinaryCritic(critic_field="usernames", weight=0.1),
                BinaryCritic(critic_field="emails", weight=0.1),
            ],
        )

    no_arguments_user_messages_by_conversation_id = [
        "Get the history of the conversation with id '1234567890'",
        "Get the history of the conversation with id '1234567890'",
        "list the messages in the conversation with id '1234567890'",
        "list the messages in the conversation with id '1234567890'",
    ]

    for user_message in no_arguments_user_messages_by_conversation_id:
        suite.add_case(
            name=f"Get conversation history by id: '{user_message}'",
            user_message=user_message,
            expected_tool_calls=[
                ExpectedToolCall(
                    func=get_messages,
                    args={
                        "conversation_id": "1234567890",
                        "channel_name": None,
                        "user_ids": None,
                        "usernames": None,
                        "emails": None,
                    },
                ),
            ],
            critics=[
                BinaryCritic(critic_field="conversation_id", weight=0.6),
                BinaryCritic(critic_field="channel_name", weight=0.1),
                BinaryCritic(critic_field="user_ids", weight=0.1),
                BinaryCritic(critic_field="usernames", weight=0.1),
                BinaryCritic(critic_field="emails", weight=0.1),
            ],
        )

    suite.add_case(
        name="Get conversation history with limit by name",
        user_message="Get the last 10 messages in the #general channel",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages,
                args={
                    "conversation_id": None,
                    "channel_name": "general",
                    "user_ids": None,
                    "usernames": None,
                    "emails": None,
                    "limit": 10,
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=0.1),
            BinaryCritic(critic_field="channel_name", weight=0.3),
            BinaryCritic(critic_field="user_ids", weight=0.1),
            BinaryCritic(critic_field="usernames", weight=0.1),
            BinaryCritic(critic_field="emails", weight=0.1),
            BinaryCritic(critic_field="limit", weight=0.3),
        ],
    )

    suite.add_case(
        name="Get conversation history with limit by id",
        user_message="Get the last 25 messages in the conversation with id '1234567890'",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages,
                args={
                    "conversation_id": "1234567890",
                    "channel_name": None,
                    "user_ids": None,
                    "usernames": None,
                    "emails": None,
                    "limit": 25,
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=0.3),
            BinaryCritic(critic_field="channel_name", weight=0.1),
            BinaryCritic(critic_field="user_ids", weight=0.1),
            BinaryCritic(critic_field="usernames", weight=0.1),
            BinaryCritic(critic_field="emails", weight=0.1),
            BinaryCritic(critic_field="limit", weight=0.3),
        ],
    )

    # Relative time eval cases by id

    suite.add_case(
        name="Get conversation history oldest relative by id (2 days ago)",
        user_message="Get the messages in the conversation with id '1234567890' starting 2 days ago",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages,
                args={
                    "conversation_id": "1234567890",
                    "channel_name": None,
                    "user_ids": None,
                    "usernames": None,
                    "emails": None,
                    "oldest_relative": "02:00:00",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=0.3),
            BinaryCritic(critic_field="channel_name", weight=0.1),
            BinaryCritic(critic_field="user_ids", weight=0.1),
            BinaryCritic(critic_field="usernames", weight=0.1),
            BinaryCritic(critic_field="emails", weight=0.1),
            RelativeTimeBinaryCritic(critic_field="oldest_relative", weight=0.3),
        ],
    )

    suite.add_case(
        name="Get conversation history oldest and latest relative by id",
        user_message="Get the messages in the conversation with id '1234567890' from 2 days ago to 3 hours ago",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages,
                args={
                    "conversation_id": "1234567890",
                    "channel_name": None,
                    "user_ids": None,
                    "usernames": None,
                    "emails": None,
                    "oldest_relative": "02:00:00",
                    "latest_relative": "00:03:00",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=0.2),
            BinaryCritic(critic_field="channel_name", weight=0.1),
            BinaryCritic(critic_field="user_ids", weight=0.1),
            BinaryCritic(critic_field="usernames", weight=0.1),
            BinaryCritic(critic_field="emails", weight=0.1),
            RelativeTimeBinaryCritic(critic_field="oldest_relative", weight=0.2),
            RelativeTimeBinaryCritic(critic_field="latest_relative", weight=0.2),
        ],
    )

    suite.add_case(
        name="Get conversation history oldest relative by id (1 week ago)",
        user_message="Get the messages in the conversation with id '1234567890' starting 1 week ago",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages,
                args={
                    "conversation_id": "1234567890",
                    "channel_name": None,
                    "user_ids": None,
                    "usernames": None,
                    "emails": None,
                    "oldest_relative": "07:00:00",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=0.3),
            BinaryCritic(critic_field="channel_name", weight=0.1),
            BinaryCritic(critic_field="user_ids", weight=0.1),
            BinaryCritic(critic_field="usernames", weight=0.1),
            BinaryCritic(critic_field="emails", weight=0.1),
            RelativeTimeBinaryCritic(critic_field="oldest_relative", weight=0.3),
        ],
    )

    suite.add_case(
        name="Get conversation history oldest relative by id (yesterday)",
        user_message="Get the messages in the conversation with id '1234567890' from yesterday",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages,
                args={
                    "conversation_id": "1234567890",
                    "channel_name": None,
                    "user_ids": None,
                    "usernames": None,
                    "emails": None,
                    "oldest_relative": "01:00:00",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=0.3),
            BinaryCritic(critic_field="channel_name", weight=0.1),
            BinaryCritic(critic_field="user_ids", weight=0.1),
            BinaryCritic(critic_field="usernames", weight=0.1),
            BinaryCritic(critic_field="emails", weight=0.1),
            RelativeTimeBinaryCritic(critic_field="oldest_relative", weight=0.3),
        ],
    )

    # Relative time eval cases by name

    suite.add_case(
        name="Get conversation history oldest relative by name (2 days ago)",
        user_message="Get the messages in the #general channel starting 2 days ago",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages,
                args={
                    "conversation_id": None,
                    "channel_name": "general",
                    "user_ids": None,
                    "usernames": None,
                    "emails": None,
                    "oldest_relative": "02:00:00",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=0.1),
            BinaryCritic(critic_field="channel_name", weight=0.3),
            BinaryCritic(critic_field="user_ids", weight=0.1),
            BinaryCritic(critic_field="usernames", weight=0.1),
            BinaryCritic(critic_field="emails", weight=0.1),
            RelativeTimeBinaryCritic(critic_field="oldest_relative", weight=0.3),
        ],
    )

    suite.add_case(
        name="Get conversation history oldest and latest relative by name",
        user_message="Get the messages in the #general channel from 2 days ago to 3 hours ago",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages,
                args={
                    "conversation_id": None,
                    "channel_name": "general",
                    "user_ids": None,
                    "usernames": None,
                    "emails": None,
                    "oldest_relative": "02:00:00",
                    "latest_relative": "00:03:00",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=0.1),
            BinaryCritic(critic_field="channel_name", weight=0.2),
            BinaryCritic(critic_field="user_ids", weight=0.1),
            BinaryCritic(critic_field="usernames", weight=0.1),
            BinaryCritic(critic_field="emails", weight=0.1),
            RelativeTimeBinaryCritic(critic_field="oldest_relative", weight=0.2),
            RelativeTimeBinaryCritic(critic_field="latest_relative", weight=0.2),
        ],
    )

    suite.add_case(
        name="Get conversation history oldest relative by name (yesterday)",
        user_message="Get the messages in the #general channel from yesterday",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages,
                args={
                    "conversation_id": None,
                    "channel_name": "general",
                    "user_ids": None,
                    "usernames": None,
                    "emails": None,
                    "oldest_relative": "01:00:00",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=0.1),
            BinaryCritic(critic_field="channel_name", weight=0.3),
            BinaryCritic(critic_field="user_ids", weight=0.1),
            BinaryCritic(critic_field="usernames", weight=0.1),
            BinaryCritic(critic_field="emails", weight=0.1),
            RelativeTimeBinaryCritic(critic_field="oldest_relative", weight=0.3),
        ],
    )

    suite.add_case(
        name="Get conversation history oldest relative by name (last week)",
        user_message="Get the messages in the #general channel from last week",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages,
                args={
                    "conversation_id": None,
                    "channel_name": "general",
                    "user_ids": None,
                    "usernames": None,
                    "emails": None,
                    "oldest_relative": "07:00:00",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=0.1),
            BinaryCritic(critic_field="channel_name", weight=0.3),
            BinaryCritic(critic_field="user_ids", weight=0.1),
            BinaryCritic(critic_field="usernames", weight=0.1),
            BinaryCritic(critic_field="emails", weight=0.1),
            RelativeTimeBinaryCritic(critic_field="oldest_relative", weight=0.3),
        ],
    )

    # Absolute time eval cases by id

    suite.add_case(
        name="Get conversation history oldest absolute by id (on a specific date)",
        user_message="Get the messages in the conversation with id '1234567890' from 2025-01-20",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages,
                args={
                    "conversation_id": "1234567890",
                    "channel_name": None,
                    "user_ids": None,
                    "usernames": None,
                    "emails": None,
                    "oldest_datetime": "2025-01-20 00:00:00",
                    "latest_datetime": "2025-01-20 23:59:59",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=0.2),
            BinaryCritic(critic_field="channel_name", weight=0.1),
            BinaryCritic(critic_field="user_ids", weight=0.1),
            BinaryCritic(critic_field="usernames", weight=0.1),
            BinaryCritic(critic_field="emails", weight=0.1),
            DatetimeCritic(
                critic_field="oldest_datetime", weight=0.2, max_difference=timedelta(minutes=2)
            ),
            DatetimeCritic(
                critic_field="latest_datetime", weight=0.2, max_difference=timedelta(minutes=2)
            ),
        ],
    )

    suite.add_case(
        name="Get conversation history oldest absolute by id (between a date range)",
        user_message="Get the messages in the conversation with id '1234567890' from 2025-01-20 to 2025-01-25",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages,
                args={
                    "conversation_id": "1234567890",
                    "channel_name": None,
                    "user_ids": None,
                    "usernames": None,
                    "emails": None,
                    "oldest_datetime": "2025-01-20 00:00:00",
                    "latest_datetime": "2025-01-25 23:59:59",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=0.2),
            BinaryCritic(critic_field="channel_name", weight=0.1),
            BinaryCritic(critic_field="user_ids", weight=0.1),
            BinaryCritic(critic_field="usernames", weight=0.1),
            BinaryCritic(critic_field="emails", weight=0.1),
            DatetimeCritic(
                critic_field="oldest_datetime", weight=0.2, max_difference=timedelta(minutes=2)
            ),
            DatetimeCritic(
                critic_field="latest_datetime", weight=0.2, max_difference=timedelta(minutes=2)
            ),
        ],
    )

    suite.add_case(
        name="Get conversation history oldest absolute by name (on a specific date)",
        user_message="Get the messages in the #general channel from 2025-01-20",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages,
                args={
                    "conversation_id": None,
                    "channel_name": "general",
                    "user_ids": None,
                    "usernames": None,
                    "emails": None,
                    "oldest_datetime": "2025-01-20 00:00:00",
                    "latest_datetime": "2025-01-20 23:59:59",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=0.1),
            BinaryCritic(critic_field="channel_name", weight=0.2),
            BinaryCritic(critic_field="user_ids", weight=0.1),
            BinaryCritic(critic_field="usernames", weight=0.1),
            BinaryCritic(critic_field="emails", weight=0.1),
            # We use a timedelta of 10 seconds because sometimes the LLM will select the limit
            # date at 23:59:59, other times it'll select the next day at 00:00:00.
            DatetimeCritic(
                critic_field="oldest_datetime", weight=0.2, max_difference=timedelta(seconds=10)
            ),
            DatetimeCritic(
                critic_field="latest_datetime", weight=0.2, max_difference=timedelta(seconds=10)
            ),
        ],
    )

    suite.add_case(
        name="Get conversation history oldest absolute by name (between a date range)",
        user_message="Get the messages in the #general channel from 2025-01-20 to 2025-01-25",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages,
                args={
                    "conversation_id": None,
                    "channel_name": "general",
                    "user_ids": None,
                    "usernames": None,
                    "emails": None,
                    "oldest_datetime": "2025-01-20 00:00:00",
                    "latest_datetime": "2025-01-25 23:59:59",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=0.1),
            BinaryCritic(critic_field="channel_name", weight=0.2),
            BinaryCritic(critic_field="user_ids", weight=0.1),
            BinaryCritic(critic_field="usernames", weight=0.1),
            BinaryCritic(critic_field="emails", weight=0.1),
            # We use a timedelta of 10 seconds because sometimes the LLM will select the limit
            # date at 23:59:59, other times it'll select the next day at 00:00:00.
            DatetimeCritic(
                critic_field="oldest_datetime", weight=0.2, max_difference=timedelta(seconds=10)
            ),
            DatetimeCritic(
                critic_field="latest_datetime", weight=0.2, max_difference=timedelta(seconds=10)
            ),
        ],
    )

    # Eval case for pagination

    suite.add_case(
        name="Get conversation history with pagination",
        user_message="get the next 5 messages",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages,
                args={
                    "conversation_id": None,
                    "channel_name": "general",
                    "user_ids": None,
                    "usernames": None,
                    "emails": None,
                    "limit": 5,
                    "next_cursor": "dXNlcjpVsDjzOTZGVDlQRA==",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=0.1),
            BinaryCritic(critic_field="channel_name", weight=0.2),
            BinaryCritic(critic_field="user_ids", weight=0.1),
            BinaryCritic(critic_field="usernames", weight=0.1),
            BinaryCritic(critic_field="emails", weight=0.1),
            BinaryCritic(critic_field="next_cursor", weight=0.2),
            BinaryCritic(critic_field="limit", weight=0.2),
        ],
        additional_messages=[
            {"role": "user", "content": "Get the last 2 messages on the general channel"},
            {
                "role": "assistant",
                "content": "",
                "tool_calls": [
                    {
                        "id": "call_1",
                        "type": "function",
                        "function": {
                            "name": "Slack_GetConversationHistoryByName",
                            "arguments": json.dumps({
                                "conversation_name": "general",
                                "limit": 2,
                            }),
                        },
                    }
                ],
            },
            {
                "role": "tool",
                "content": json.dumps({
                    "messages": [
                        {
                            "blocks": [
                                {
                                    "block_id": "abc123",
                                    "elements": [
                                        {
                                            "elements": [
                                                {
                                                    "text": "Almost there, Boss, need to get some evals in!",
                                                    "type": "text",
                                                }
                                            ],
                                            "type": "rich_text_section",
                                        }
                                    ],
                                    "type": "rich_text",
                                }
                            ],
                            "client_msg_id": "msg_id_0",
                            "datetime_timestamp": "2025-01-21 16:59:55",
                            "team": "617263616465207465616D20697320617420626F7373206C6576656C",
                            "text": "Almost there, Boss, need to get some evals in!",
                            "ts": "1737507595.598529",
                            "type": "message",
                            "user": "77686F2069732074686520626F73733F",
                        },
                        {
                            "blocks": [
                                {
                                    "block_id": "xyz456",
                                    "elements": [
                                        {
                                            "elements": [
                                                {
                                                    "text": "hey, are the Slack Tools ready yet?",
                                                    "type": "text",
                                                }
                                            ],
                                            "type": "rich_text_section",
                                        }
                                    ],
                                    "type": "rich_text",
                                }
                            ],
                            "client_msg_id": "msg_id_1",
                            "datetime_timestamp": "2025-01-21 16:57:35",
                            "team": "617263616465207465616D20697320617420626F7373206C6576656C",
                            "text": "hey, are the Slack Tools ready yet?",
                            "ts": "1737507595.598529",
                            "type": "message",
                            "user": "73616D2069732074686520626F7373",
                        },
                    ],
                    "next_cursor": "dXNlcjpVsDjzOTZGVDlQRA==",
                }),
                "tool_call_id": "call_1",
                "name": "Slack_GetConversationHistoryByName",
            },
            {
                "role": "assistant",
                "content": 'Here are the last 2 messages from the general channel:\n\n1. **User:** 77686F2069732074686520626F73733F  \n   **Message:** "Almost there, Boss, need to get some evals in!"  \n   **Timestamp:** 2025-01-21 16:59:55\n\n2. **User:** 73616D2069732074686520626F7373  \n   **Message:** "hey, are the Slack Tools ready yet?"  \n   **Timestamp:** 2025-01-21 16:57:35',
            },
        ],
    )

    return suite



import ast

import pytest
from arcade_core.parse import get_tools_from_ast


@pytest.mark.parametrize(
    "source, expected_tools",
    [
        pytest.param(
            """
@tool
def my_function():
    pass
    """,
            ["my_function"],
            id="function with tool decorator",
        ),
        pytest.param(
            """
import arcade.sdk as arc
@arc.tool
def another_function():
    pass
    """,
            ["another_function"],
            id="function with arc.tool decorator",
        ),
        pytest.param(
            """
def no_decorator_function():
    pass
    """,
            [],
            id="function without decorator",
        ),
        pytest.param(
            """
@other_decorator
def different_function():
    pass
    """,
            [],
            id="function with other decorator",
        ),
    ],
)
def test_get_function_name_if_decorated(source, expected_tools):
    tree = ast.parse(source)
    tools = get_tools_from_ast(tree)
    assert tools == expected_tools



from typing import Annotated

import httpx
from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import X
from arcade_tdk.errors import RetryableToolError

from arcade_x.tools.utils import (
    expand_urls_in_user_description,
    expand_urls_in_user_url,
    get_headers_with_token,
)

# Users Lookup Tools. See developer docs for additional available query parameters:
# https://developer.x.com/en/docs/x-api/users/lookup/api-reference


@tool(requires_auth=X(scopes=["users.read", "tweet.read"]))
async def lookup_single_user_by_username(
    context: ToolContext,
    username: Annotated[str, "The username of the X (Twitter) user to look up"],
) -> Annotated[dict, "User information including id, name, username, and description"]:
    """Look up a user on X (Twitter) by their username."""

    headers = get_headers_with_token(context)

    user_fields = ",".join([
        "created_at",
        "description",
        "id",
        "location",
        "most_recent_tweet_id",
        "name",
        "pinned_tweet_id",
        "profile_image_url",
        "protected",
        "public_metrics",
        "url",
        "username",
        "verified",
        "verified_type",
        "withheld",
        "entities",
    ])
    url = f"https://api.x.com/2/users/by/username/{username}?user.fields={user_fields}"

    async with httpx.AsyncClient() as client:
        response = await client.get(url, headers=headers, timeout=10)
        if response.status_code == 404:
            # User not found
            raise RetryableToolError(
                "User not found",
                developer_message=f"User with username '{username}' not found.",
                additional_prompt_content="Please check the username and try again.",
                retry_after_ms=500,  # Play nice with X API rate limits
            )
        response.raise_for_status()
    # Parse the response JSON
    user_data = response.json()["data"]

    user_data = expand_urls_in_user_description(user_data, delete_entities=False)
    user_data = expand_urls_in_user_url(user_data, delete_entities=True)

    return {"data": user_data}



from arcade_firecrawl.tools import (
    cancel_crawl,
    crawl_website,
    get_crawl_data,
    get_crawl_status,
    map_website,
    scrape_url,
)

__all__ = [
    "cancel_crawl",
    "crawl_website",
    "get_crawl_data",
    "get_crawl_status",
    "map_website",
    "scrape_url",
]



from unittest.mock import AsyncMock, patch

import pytest
from arcade_tdk.errors import RetryableToolError, ToolExecutionError
from httpx import Response

from arcade_github.tools.models import (
    DiffSide,
    ReviewCommentSubjectType,
)
from arcade_github.tools.pull_requests import (
    create_reply_for_review_comment,
    create_review_comment,
    get_pull_request,
    list_pull_request_commits,
    list_pull_requests,
    list_review_comments_on_pull_request,
    update_pull_request,
)


@pytest.fixture
def mock_context():
    context = AsyncMock()
    context.authorization.token = "mock_token"  # noqa: S105
    return context


@pytest.fixture
def mock_client():
    with patch("arcade_github.tools.pull_requests.httpx.AsyncClient") as client:
        yield client.return_value.__aenter__.return_value


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "func,args,status_code,json_response,expected_result,error_message",
    [
        (list_pull_requests, ("owner", "repo"), 200, [], '{"pull_requests": []}', None),
        (
            get_pull_request,
            ("owner", "repo", 1),
            404,
            {"message": "Not Found"},
            None,
            "Error accessing.*: Resource not found",
        ),
        (
            update_pull_request,
            ("owner", "repo", 1, "New Title"),
            409,
            {"message": "Conflict"},
            None,
            "Error accessing.*: Failed to process request",
        ),
        (
            list_pull_request_commits,
            ("owner", "repo", 1),
            500,
            {"message": "Internal Server Error"},
            None,
            "Error accessing.*: Failed to process request",
        ),
        (
            list_review_comments_on_pull_request,
            ("owner", "repo", 1),
            403,
            {"message": "API rate limit exceeded"},
            None,
            "Error accessing.*: Forbidden",
        ),
    ],
)
async def test_pull_request_functions(
    mock_context,
    mock_client,
    func,
    args,
    status_code,
    json_response,
    expected_result,
    error_message,
):
    mock_client.get.return_value = mock_client.post.return_value = (
        mock_client.patch.return_value
    ) = Response(status_code, json=json_response)

    if error_message:
        with pytest.raises(ToolExecutionError, match=error_message):
            await func(mock_context, *args)
    else:
        result = await func(mock_context, *args)
        assert result == expected_result


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "func,args,json_response,expected_assertions",
    [
        (
            list_pull_requests,
            ("owner", "repo"),
            [
                {
                    "number": 1,
                    "title": "Test PR",
                    "body": "This is a test PR",
                    "state": "open",
                    "html_url": "https://github.com/owner/repo/pull/1",
                    "created_at": "2023-05-01T12:00:00Z",
                    "updated_at": "2023-05-01T12:00:00Z",
                    "user": {"login": "testuser"},
                    "base": {"ref": "main"},
                    "head": {"ref": "feature-branch"},
                }
            ],
            ["Test PR", "https://github.com/owner/repo/pull/1"],
        ),
        (
            update_pull_request,
            ("owner", "repo", 1, "Updated PR Title", "Updated PR body"),
            {
                "number": 1,
                "title": "Updated PR Title",
                "body": "Updated PR body",
                "state": "open",
                "html_url": "https://github.com/owner/repo/pull/1",
                "created_at": "2023-05-01T12:00:00Z",
                "updated_at": "2023-05-02T12:00:00Z",
                "user": {"login": "testuser"},
            },
            ["Updated PR Title", "Updated PR body"],
        ),
        (
            list_pull_request_commits,
            ("owner", "repo", 1),
            [
                {
                    "sha": "6dcb09b5b57875f334f61aebed695e2e4193db5e",
                    "commit": {
                        "author": {
                            "name": "Test Author",
                            "email": "author@example.com",
                            "date": "2023-05-01T12:00:00Z",
                        },
                        "message": "Test commit message",
                    },
                }
            ],
            ["6dcb09b5b57875f334f61aebed695e2e4193db5e", "Test commit message"],
        ),
        (
            create_reply_for_review_comment,
            ("owner", "repo", 1, 42, "Thanks for the suggestion."),
            {
                "id": 123,
                "body": "Thanks for the suggestion.",
                "user": {"login": "testuser"},
                "created_at": "2023-05-02T12:00:00Z",
                "updated_at": "2023-05-02T12:00:00Z",
            },
            ["Thanks for the suggestion.", "testuser"],
        ),
        (
            list_review_comments_on_pull_request,
            ("owner", "repo", 1),
            [
                {
                    "id": 1,
                    "body": "Great changes!",
                    "user": {"login": "reviewer1"},
                    "created_at": "2023-05-01T12:00:00Z",
                    "updated_at": "2023-05-01T12:00:00Z",
                    "path": "file1.txt",
                    "line": 5,
                }
            ],
            ["Great changes!", "reviewer1", "file1.txt"],
        ),
        (
            get_pull_request,
            ("owner", "repo", 1, False, False),
            {
                "number": 1,
                "title": "Test PR",
                "body": "This is a test PR",
                "state": "open",
                "html_url": "https://github.com/owner/repo/pull/1",
                "created_at": "2023-05-01T12:00:00Z",
                "updated_at": "2023-05-01T12:00:00Z",
                "user": {"login": "testuser"},
                "base": {"ref": "main"},
                "head": {"ref": "feature-branch"},
            },
            ["Test PR", "https://github.com/owner/repo/pull/1"],
        ),
        (
            get_pull_request,
            ("owner", "repo", 1, True, False),
            {
                "number": 1,
                "title": "Test PR",
                "body": "This is a test PR",
                "state": "open",
                "html_url": "https://github.com/owner/repo/pull/1",
                "created_at": "2023-05-01T12:00:00Z",
                "updated_at": "2023-05-01T12:00:00Z",
                "user": {"login": "testuser"},
                "base": {"ref": "main"},
                "head": {"ref": "feature-branch"},
                "diff_content": "Sample diff content",
            },
            ["Test PR", "https://github.com/owner/repo/pull/1", "diff_content"],
        ),
        (
            create_review_comment,
            (
                "owner",
                "repo",
                1,
                "Great changes!",
                "file1.txt",
                "6dcb09b5b57875f334f61aebed695e2e4193db5e",
                1,
                2,
                DiffSide.RIGHT,
                None,
                ReviewCommentSubjectType.LINE,
            ),
            {
                "id": 1,
                "body": "Great changes!",
                "path": "file1.txt",
                "line": 2,
                "side": "RIGHT",
                "commit_id": "6dcb09b5b57875f334f61aebed695e2e4193db5e",
                "user": {"login": "testuser"},
                "created_at": "2023-05-01T12:00:00Z",
                "updated_at": "2023-05-01T12:00:00Z",
                "html_url": "https://github.com/owner/repo/pull/1#discussion_r1",
            },
            ["Great changes!", "file1.txt", "6dcb09b5b57875f334f61aebed695e2e4193db5e"],
        ),
    ],
)
async def test_pull_request_functions_success(
    mock_context, mock_client, func, args, json_response, expected_assertions
):
    mock_client.get.return_value = mock_client.post.return_value = (
        mock_client.patch.return_value
    ) = Response(200, json=json_response)

    result = await func(mock_context, *args)
    for assertion in expected_assertions:
        assert assertion in result


@pytest.mark.asyncio
async def test_create_review_comment_file_subject_type(mock_context, mock_client):
    mock_client.post.return_value = Response(
        200,
        json={
            "id": 1,
            "body": "File comment",
            "path": "file1.txt",
            "commit_id": "6dcb09b5b57875f334f61aebed695e2e4193db5e",
            "user": {"login": "testuser"},
            "created_at": "2023-05-01T12:00:00Z",
            "updated_at": "2023-05-01T12:00:00Z",
            "html_url": "https://github.com/owner/repo/pull/1#discussion_r1",
        },
    )

    result = await create_review_comment(
        mock_context,
        "owner",
        "repo",
        1,
        "File comment",
        "file1.txt",
        "6dcb09b5b57875f334f61aebed695e2e4193db5e",
        subject_type=ReviewCommentSubjectType.FILE,
    )

    assert "File comment" in result
    assert "file1.txt" in result
    assert "6dcb09b5b57875f334f61aebed695e2e4193db5e" in result
    assert "start_line" not in mock_client.post.call_args[1]["json"]
    assert "end_line" not in mock_client.post.call_args[1]["json"]


@pytest.mark.asyncio
async def test_create_review_comment_missing_commit_id(mock_context, mock_client):
    mock_client.get.return_value = Response(
        200,
        json=[{"sha": "latest_commit_sha"}],
    )
    mock_client.post.return_value = Response(
        200,
        json={
            "id": 1,
            "body": "Comment with auto-fetched commit ID",
            "path": "file1.txt",
            "commit_id": "latest_commit_sha",
            "user": {"login": "testuser"},
            "created_at": "2023-05-01T12:00:00Z",
            "updated_at": "2023-05-01T12:00:00Z",
            "html_url": "https://github.com/owner/repo/pull/1#discussion_r1",
        },
    )

    result = await create_review_comment(
        mock_context,
        "owner",
        "repo",
        1,
        "Comment with auto-fetched commit ID",
        "file1.txt",
        start_line=1,
        end_line=2,
    )

    assert "Comment with auto-fetched commit ID" in result
    assert "latest_commit_sha" in result
    assert mock_client.get.called
    assert mock_client.post.called


@pytest.mark.asyncio
async def test_create_review_comment_invalid_input(mock_context, mock_client):
    with pytest.raises(
        RetryableToolError, match="'start_line' and 'end_line' parameters are required"
    ):
        await create_review_comment(
            mock_context,
            "owner",
            "repo",
            1,
            "Invalid comment",
            "file1.txt",
            subject_type=ReviewCommentSubjectType.LINE,
        )


@pytest.mark.asyncio
async def test_create_review_comment_no_commits(mock_context, mock_client):
    mock_client.get.return_value = Response(200, json=[])

    with pytest.raises(RetryableToolError, match="Failed to get the latest commit SHA"):
        await create_review_comment(
            mock_context,
            "owner",
            "repo",
            1,
            "Comment with no commits",
            "file1.txt",
            start_line=1,
            end_line=2,
        )



from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_slack
from arcade_slack.tools.chat import get_conversation_metadata

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.8,
    warn_threshold=0.9,
)


catalog = ToolCatalog()
# Register the Slack tools
catalog.add_module(arcade_slack)


@tool_eval()
def get_conversations_metadata_eval_suite() -> EvalSuite:
    """Create an evaluation suite for tools getting conversations metadata."""
    suite = EvalSuite(
        name="Slack Tools Evaluation",
        system_message="You are an AI assistant that can interact with Slack to get information from conversations, users, etc.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Get channel metadata by name",
        user_message="Get the metadata of the #general channel",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_conversation_metadata,
                args={
                    "conversation_id": None,
                    "channel_name": "general",
                    "user_ids": None,
                    "usernames": None,
                    "emails": None,
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=1 / 5),
            BinaryCritic(critic_field="channel_name", weight=1 / 5),
            BinaryCritic(critic_field="user_ids", weight=1 / 5),
            BinaryCritic(critic_field="usernames", weight=1 / 5),
            BinaryCritic(critic_field="emails", weight=1 / 5),
        ],
    )

    suite.add_case(
        name="Get conversation metadata by id",
        user_message="Get the metadata of the conversation with id '1234567890'",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_conversation_metadata,
                args={
                    "conversation_id": "1234567890",
                    "channel_name": None,
                    "user_ids": None,
                    "usernames": None,
                    "emails": None,
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=1 / 5),
            BinaryCritic(critic_field="channel_name", weight=1 / 5),
            BinaryCritic(critic_field="user_ids", weight=1 / 5),
            BinaryCritic(critic_field="usernames", weight=1 / 5),
            BinaryCritic(critic_field="emails", weight=1 / 5),
        ],
    )

    suite.add_case(
        name="Get conversation metadata by username mentioning DM",
        user_message="get the metadata of the DM with janedoe",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_conversation_metadata,
                args={
                    "conversation_id": None,
                    "channel_name": None,
                    "user_ids": None,
                    "usernames": ["janedoe"],
                    "emails": None,
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=1 / 5),
            BinaryCritic(critic_field="channel_name", weight=1 / 5),
            BinaryCritic(critic_field="user_ids", weight=1 / 5),
            BinaryCritic(critic_field="usernames", weight=1 / 5),
            BinaryCritic(critic_field="emails", weight=1 / 5),
        ],
    )

    suite.add_case(
        name="Get conversation metadata by username mentioning IM",
        user_message="get metadata about my IM with janedoe",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_conversation_metadata,
                args={
                    "conversation_id": None,
                    "channel_name": None,
                    "user_ids": None,
                    "usernames": ["janedoe"],
                    "emails": None,
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=1 / 5),
            BinaryCritic(critic_field="channel_name", weight=1 / 5),
            BinaryCritic(critic_field="user_ids", weight=1 / 5),
            BinaryCritic(critic_field="usernames", weight=1 / 5),
            BinaryCritic(critic_field="emails", weight=1 / 5),
        ],
    )

    suite.add_case(
        name="Get conversation metadata by email mentioning DM",
        user_message="get the metadata of the DM with jane.doe@acme.com",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_conversation_metadata,
                args={
                    "conversation_id": None,
                    "channel_name": None,
                    "user_ids": None,
                    "usernames": None,
                    "emails": ["jane.doe@acme.com"],
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=1 / 5),
            BinaryCritic(critic_field="channel_name", weight=1 / 5),
            BinaryCritic(critic_field="user_ids", weight=1 / 5),
            BinaryCritic(critic_field="usernames", weight=1 / 5),
            BinaryCritic(critic_field="emails", weight=1 / 5),
        ],
    )

    suite.add_case(
        name="Get conversation metadata by email mentioning IM",
        user_message="get the metadata of the IM with jane.doe@acme.com",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_conversation_metadata,
                args={
                    "conversation_id": None,
                    "channel_name": None,
                    "user_ids": None,
                    "usernames": None,
                    "emails": ["jane.doe@acme.com"],
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=1 / 5),
            BinaryCritic(critic_field="channel_name", weight=1 / 5),
            BinaryCritic(critic_field="user_ids", weight=1 / 5),
            BinaryCritic(critic_field="usernames", weight=1 / 5),
            BinaryCritic(critic_field="emails", weight=1 / 5),
        ],
    )

    suite.add_case(
        name="Get conversation metadata by mixed user ID, email, and username",
        user_message=(
            "get the metadata of the multi-person conversation I have with these users together: "
            "janedoe, john@acme.com, and U0123456789"
        ),
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_conversation_metadata,
                args={
                    "conversation_id": None,
                    "channel_name": None,
                    "user_ids": ["U0123456789"],
                    "usernames": ["janedoe"],
                    "emails": ["john@acme.com"],
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=1 / 5),
            BinaryCritic(critic_field="channel_name", weight=1 / 5),
            BinaryCritic(critic_field="user_ids", weight=1 / 5),
            BinaryCritic(critic_field="usernames", weight=1 / 5),
            BinaryCritic(critic_field="emails", weight=1 / 5),
        ],
    )

    return suite



import asyncio

import pytest
from arcade_serve.mcp.message_processor import MCPMessageProcessor, create_message_processor
from arcade_serve.mcp.types import InitializeRequest, PingRequest


@pytest.mark.asyncio
async def test_message_processor_parses_initialize_json():
    """Ensure JSON initialize strings are converted into InitializeRequest objects."""
    json_init = '{"jsonrpc":"2.0","id":1,"method":"initialize","params":{}}\n'
    processor = MCPMessageProcessor()

    result = await processor.process_request(json_init)

    assert isinstance(result, InitializeRequest)
    assert result.id == 1
    assert result.method == "initialize"


@pytest.mark.asyncio
async def test_message_processor_passes_notifications_unchanged():
    """Unknown notifications should be passed through as parsed dictionaries without errors."""
    json_notification = '{"jsonrpc":"2.0","id":null,"method":"notifications/custom","params":{}}\n'
    processor = MCPMessageProcessor()

    result = await processor.process_request(json_notification)

    # The MCPMessageProcessor keeps unknown notifications as simple dicts
    assert isinstance(result, dict)
    assert result["method"] == "notifications/custom"


@pytest.mark.asyncio
async def test_message_processor_middleware_execution_order(monkeypatch):
    """Middleware (sync + async) should be executed in the order they were added."""

    order: list[str] = []

    def mw_sync(msg, direction):  # type: ignore[return-value]
        order.append("sync")
        return msg

    async def mw_async(msg, direction):  # type: ignore[return-value]
        await asyncio.sleep(0)  # ensure it is truly async
        order.append("async")
        return msg

    processor = create_message_processor(mw_sync, mw_async)

    # Use a pre-parsed PingRequest instance so we don't test parsing again here
    ping = PingRequest(id=42)

    _ = await processor.process_request(ping)

    assert order == ["sync", "async"]



import os

from arcade_google_maps.enums import GoogleMapsDistanceUnit, GoogleMapsTravelMode

DEFAULT_GOOGLE_LANGUAGE = os.getenv("ARCADE_GOOGLE_LANGUAGE", "en")

DEFAULT_GOOGLE_MAPS_LANGUAGE = os.getenv("ARCADE_GOOGLE_MAPS_LANGUAGE", DEFAULT_GOOGLE_LANGUAGE)
DEFAULT_GOOGLE_MAPS_COUNTRY = os.getenv("ARCADE_GOOGLE_MAPS_COUNTRY")
DEFAULT_GOOGLE_MAPS_DISTANCE_UNIT = GoogleMapsDistanceUnit(
    os.getenv("ARCADE_GOOGLE_MAPS_DISTANCE_UNIT", GoogleMapsDistanceUnit.KM.value)
)
DEFAULT_GOOGLE_MAPS_TRAVEL_MODE = GoogleMapsTravelMode(
    os.getenv("ARCADE_GOOGLE_MAPS_TRAVEL_MODE", GoogleMapsTravelMode.BEST.value)
)



from enum import Enum


class SubredditListingType(str, Enum):
    HOT = "hot"
    NEW = "new"
    RISING = "rising"
    TOP = "top"  # time-based
    CONTROVERSIAL = "controversial"  # time-based

    def is_time_based(self) -> bool:
        return self in [SubredditListingType.TOP, SubredditListingType.CONTROVERSIAL]


class RedditTimeFilter(str, Enum):
    NOW = "NOW"
    TODAY = "TODAY"
    THIS_WEEK = "THIS_WEEK"
    THIS_MONTH = "THIS_MONTH"
    THIS_YEAR = "THIS_YEAR"
    ALL_TIME = "ALL_TIME"

    def to_api_value(self) -> str:
        _map = {
            RedditTimeFilter.NOW: "hour",
            RedditTimeFilter.TODAY: "day",
            RedditTimeFilter.THIS_WEEK: "week",
            RedditTimeFilter.THIS_MONTH: "month",
            RedditTimeFilter.THIS_YEAR: "year",
            RedditTimeFilter.ALL_TIME: "all",
        }
        return _map[self]


class RedditThingType(str, Enum):
    """The type of a Reddit 'thing'.

    Typically used as a prefix for fullnames, e.g. t1_1234567890
    is the fullname of a comment with id 1234567890
    """

    COMMENT = "t1"
    ACCOUNT = "t2"
    LINK = "t3"
    MESSAGE = "t4"
    SUBREDDIT = "t5"
    AWARD = "t6"



import arcade_google  # pip install arcade_google
import arcade_search  # pip install arcade_search
from arcade_core.catalog import ToolCatalog
from arcade_serve.mcp.stdio import StdioServer

# 2. Create and populate the tool catalog
catalog = ToolCatalog()
catalog.add_module(arcade_google)  # Registers all tools in the package
catalog.add_module(arcade_search)


# 3. Main entrypoint
async def main():
    # Create the worker with the tool catalog
    worker = StdioServer(catalog)

    # Run the worker
    await worker.run()


if __name__ == "__main__":
    import asyncio

    asyncio.run(main())



from typing import Annotated

from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Reddit
from arcade_tdk.errors import ToolExecutionError

from arcade_reddit.client import RedditClient
from arcade_reddit.enums import (
    RedditTimeFilter,
    SubredditListingType,
)
from arcade_reddit.utils import (
    create_fullname_for_multiple_posts,
    create_path_for_post,
    normalize_subreddit_name,
    parse_get_content_of_multiple_posts_response,
    parse_get_content_of_post_response,
    parse_get_posts_in_subreddit_response,
    parse_get_top_level_comments_response,
    parse_subreddit_rules_response,
    parse_user_posts_response,
    remove_none_values,
    resolve_subreddit_access,
)


@tool(requires_auth=Reddit(scopes=["read"]))
async def get_posts_in_subreddit(
    context: ToolContext,
    subreddit: Annotated[str, "The name of the subreddit to fetch posts from"],
    listing: Annotated[
        SubredditListingType,
        (
            "The type of listing to fetch. For simple listings such as 'hot', 'new', or 'rising', "
            "the 'time_range' parameter is ignored. For time-based listings such as "
            "'top' or 'controversial', the 'time_range' parameter is required."
        ),
    ] = SubredditListingType.HOT,
    limit: Annotated[int, "The maximum number of posts to fetch. Default is 10, max is 100."] = 10,
    cursor: Annotated[str | None, "The pagination token from a previous call"] = None,
    time_range: Annotated[
        RedditTimeFilter,
        "The time range for filtering posts. Must be provided if the listing type is "
        f"{SubredditListingType.TOP.value} or {SubredditListingType.CONTROVERSIAL.value}. "
        f"Otherwise, it is ignored. Defaults to {RedditTimeFilter.TODAY.value}.",
    ] = RedditTimeFilter.TODAY,
) -> Annotated[dict, "A dictionary with a cursor for the next page and a list of posts"]:
    """Gets posts titles, links, and other metadata in the specified subreddit

    The time_range is required if the listing type is 'top' or 'controversial'.
    """
    client = RedditClient(context.get_auth_token_or_empty())

    params = {"limit": limit, "after": cursor}
    if listing.is_time_based():
        params["t"] = time_range.to_api_value()

    params = remove_none_values(params)
    subreddit = normalize_subreddit_name(subreddit)
    data = await client.get(f"r/{subreddit}/{listing.value}", params=params)
    result = parse_get_posts_in_subreddit_response(data)

    return result


@tool(requires_auth=Reddit(scopes=["read"]))
async def get_content_of_post(
    context: ToolContext,
    post_identifier: Annotated[
        str,
        "The identifier of the Reddit post. "
        "The identifier may be a reddit URL to the post, a permalink to the post, "
        "a fullname for the post, or a post id.",
    ],
) -> Annotated[dict, "The content (body) of the Reddit post"]:
    """Get the content (body) of a Reddit post by its identifier."""
    client = RedditClient(context.get_auth_token_or_empty())

    path = create_path_for_post(post_identifier)
    data = await client.get(f"{path}.json")
    result = parse_get_content_of_post_response(data)

    return result


@tool(requires_auth=Reddit(scopes=["read"]))
async def get_content_of_multiple_posts(
    context: ToolContext,
    post_identifiers: Annotated[
        list[str],
        "A list of Reddit post identifiers. "
        "The identifiers may be reddit URLs to the posts, permalinks to the posts, "
        "fullnames for the posts, or post ids. Must be less than or equal to 100 identifiers.",
    ],
) -> Annotated[dict, "A dictionary containing the content of multiple Reddit posts"]:
    """Get the content (body) of multiple Reddit posts by their identifiers.

    Efficiently retrieve the content of multiple posts in a single request.
    Always use this tool to retrieve more than one post's content.
    """
    client = RedditClient(context.get_auth_token_or_empty())

    fullnames, warnings = create_fullname_for_multiple_posts(post_identifiers)
    data = await client.get("api/info.json", params={"id": ",".join(fullnames)})
    posts = parse_get_content_of_multiple_posts_response(data)

    return {"posts": posts, "warnings": warnings}


@tool(requires_auth=Reddit(scopes=["read"]))
async def get_top_level_comments(
    context: ToolContext,
    post_identifier: Annotated[
        str,
        "The identifier of the Reddit post to fetch comments from. "
        "The identifier may be a reddit URL, a permalink, a fullname, or a post id.",
    ],
) -> Annotated[dict, "A dictionary with a list of top level comments"]:
    """Get the first page of top-level comments of a Reddit post."""
    client = RedditClient(context.get_auth_token_or_empty())

    path = create_path_for_post(post_identifier)
    data = await client.get(f"{path}.json")
    result = parse_get_top_level_comments_response(data)

    return result


@tool(requires_auth=Reddit(scopes=["read"]))
async def check_subreddit_access(
    context: ToolContext,
    subreddit: Annotated[str, "The name of the subreddit to check access for"],
) -> Annotated[
    dict,
    "A dict indicating whether the subreddit exists and is accessible to the authenticated user",
]:
    """
    Checks whether the specified subreddit exists and also if it is accessible
    to the authenticated user.

    Returns:
        {"exists": True, "accessible": True} if the subreddit exists and is accessible.
        {"exists": True, "accessible": False} if the subreddit exists but is private or restricted.
        {"exists": False, "accessible": False} if the subreddit does not exist.
    """
    client = RedditClient(context.get_auth_token_or_empty())

    return await resolve_subreddit_access(client, subreddit)


@tool(requires_auth=Reddit(scopes=["read"]))
async def get_subreddit_rules(
    context: ToolContext,
    subreddit: Annotated[str, "The name of the subreddit for which to fetch rules"],
) -> Annotated[dict, "A dictionary containing the subreddit rules"]:
    """Gets the rules of the specified subreddit"""
    client = RedditClient(context.get_auth_token_or_empty())

    normalized_subreddit = normalize_subreddit_name(subreddit)
    data = await client.get(f"r/{normalized_subreddit}/about/rules")

    return parse_subreddit_rules_response(data)


@tool(requires_auth=Reddit(scopes=["identity"]))
async def get_my_username(context: ToolContext) -> str:
    """Get the Reddit username of the authenticated user"""
    client = RedditClient(context.get_auth_token_or_empty())
    user_info = await client.get("api/v1/me")
    username: str = user_info.get("name", "")

    if not username:
        raise ToolExecutionError(message="Failed to retrieve the authenticated user's name")

    return username


@tool(requires_auth=Reddit(scopes=["identity", "history", "read"]))
async def get_my_posts(
    context: ToolContext,
    limit: Annotated[
        int, "The maximum number of posts to fetch. Default is 10. Maximum is 100"
    ] = 10,
    include_body: Annotated[
        bool, "Whether to include the body (content) of the posts. Defaults to True."
    ] = True,
    cursor: Annotated[str | None, "The pagination token from a previous call"] = None,
) -> Annotated[
    dict,
    "A dictionary with a cursor for the next page and "
    "a list of posts created by the authenticated user",
]:
    """Get posts that were created by the authenticated user sorted by newest first"""
    client = RedditClient(context.get_auth_token_or_empty())

    username = await get_my_username(context=context)
    params = {"limit": limit, "after": cursor}
    params = remove_none_values(params)

    posts_data = await client.get(f"user/{username}/submitted", params=params)

    return await parse_user_posts_response(context, posts_data, include_body)



from arcade_notion_toolkit.tools.pages import (
    create_page,
    get_page_content_by_id,
    get_page_content_by_title,
)
from arcade_notion_toolkit.tools.search import (
    get_object_metadata,
    get_workspace_structure,
    search_by_title,
)

__all__ = [
    "create_page",
    "get_object_metadata",
    "get_page_content_by_id",
    "get_page_content_by_title",
    "search_by_title",
    "get_workspace_structure",
]



from unittest.mock import patch

import pytest

from arcade_confluence.client import ConfluenceClientV2


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "space_identifier, is_id",
    [
        (
            "12345",
            True,
        ),
        (
            "test-space",
            False,
        ),
    ],
)
async def test_get_space_id(space_identifier, is_id) -> None:
    with patch("arcade_confluence.client.ConfluenceClient._get_cloud_id", return_value=None):
        client_v2 = ConfluenceClientV2("fake-token")
        with patch(
            "arcade_confluence.client.ConfluenceClientV2.get_space_by_key",
            return_value={"space": {"id": "12345"}},
        ):
            space_id = await client_v2.get_space_id(space_identifier)
        if is_id:
            assert space_id == space_identifier
        else:
            assert space_id == "12345"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "page_identifier, is_id",
    [
        (
            "67890",
            True,
        ),
        (
            "test-page",
            False,
        ),
    ],
)
async def test_get_page_id(page_identifier, is_id) -> None:
    with patch("arcade_confluence.client.ConfluenceClient._get_cloud_id", return_value=None):
        client_v2 = ConfluenceClientV2("fake-token")
        with patch(
            "arcade_confluence.client.ConfluenceClientV2.get_page_by_title",
            return_value={"page": {"id": "67890"}},
        ):
            page_id = await client_v2.get_page_id(page_identifier)
        if is_id:
            assert page_id == page_identifier
        else:
            assert page_id == "67890"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "space_identifier, is_id",
    [
        (
            "12345",
            True,
        ),
        (
            "test-space",
            False,
        ),
    ],
)
async def test_get_space(space_identifier, is_id) -> None:
    with patch("arcade_confluence.client.ConfluenceClient._get_cloud_id", return_value=None):
        client_v2 = ConfluenceClientV2("fake-token")

        mock_space = {"space": {"id": "12345", "key": "TEST"}}

        with (
            patch(
                "arcade_confluence.client.ConfluenceClientV2.get_space_by_id",
                return_value=mock_space,
            ) as mock_by_id,
            patch(
                "arcade_confluence.client.ConfluenceClientV2.get_space_by_key",
                return_value=mock_space,
            ) as mock_by_key,
        ):
            result = await client_v2.get_space(space_identifier)

            if is_id:
                mock_by_id.assert_called_once_with(space_identifier)
                mock_by_key.assert_not_called()
            else:
                mock_by_id.assert_not_called()
                mock_by_key.assert_called_once_with(space_identifier)

            assert result == mock_space



import json
import logging
import sys
import time
from typing import Any

from arcade_serve.mcp.types import (
    JSONRPCError,
    JSONRPCRequest,
    JSONRPCResponse,
    MCPMessage,
)

logger = logging.getLogger("arcade.mcp")


class MCPLoggingMiddleware:
    """
    Middleware for logging MCP requests and responses.
    Logs request and response details, including timing and errors.
    """

    def __init__(
        self,
        log_level: str = "INFO",
        log_request_body: bool = False,
        log_response_body: bool = False,
        log_errors: bool = True,
        min_duration_to_log_ms: int = 0,
        stdio_mode: bool = False,
    ) -> None:
        """
        Initialize the MCP logging middleware.

        Args:
            log_level: Logging level (default: "INFO").
            log_request_body: Whether to log full request bodies (default: False).
            log_response_body: Whether to log full response bodies (default: False).
            log_errors: Whether to log errors at ERROR level (default: True).
            min_duration_to_log_ms: Minimum duration in ms to log (0 logs all).
            stdio_mode: Whether running in stdio mode (redirects logs to stderr).
        """
        self.log_level = getattr(logging, log_level.upper())
        self.log_request_body = log_request_body
        self.log_response_body = log_response_body
        self.log_errors = log_errors
        self.min_duration_to_log_ms = min_duration_to_log_ms
        self.request_log_format = "[MCP>] {method}{params_str} (id: {id})"
        self.response_log_format = "[MCP<] {method} completed in {duration:.2f}ms (id: {id})"
        self.error_log_format = "[MCP!] {method} error: {error} (id: {id})"

        # If in stdio mode, ensure MCP logs go to stderr
        if stdio_mode:
            self._redirect_logs_to_stderr()

        # Log that middleware is initialized
        logger.debug(f"MCP logging middleware initialized (level: {log_level})")

    def _redirect_logs_to_stderr(self) -> None:
        """Redirect MCP logs to stderr to avoid interfering with stdio communication."""
        # Remove any existing handlers
        for handler in logger.handlers[:]:
            logger.removeHandler(handler)

        # Add a stderr handler
        stderr_handler = logging.StreamHandler(sys.stderr)
        formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        stderr_handler.setFormatter(formatter)
        logger.addHandler(stderr_handler)

        # Ensure we're not propagating to root logger which might log to stdout
        logger.propagate = False

        logger.debug("MCP logs redirected to stderr for stdio mode")

    def __call__(self, message: MCPMessage, direction: str) -> MCPMessage:
        """
        Process and log an MCP message.

        Args:
            message: The MCP message to process.
            direction: The message direction ("request" or "response").

        Returns:
            The original message (unmodified).
        """
        if direction == "request":
            self._log_request(message)
        else:
            self._log_response(message)
        return message

    def _log_request(self, message: MCPMessage) -> None:
        """
        Log an MCP request message.
        """
        if not isinstance(message, JSONRPCRequest):
            logger.debug(f"Ignoring non-request message: {type(message).__name__}")
            return

        try:
            # Store request start time for duration calculation
            message._mcp_start_time = time.time()  # type: ignore[attr-defined]

            # Format parameters for logging
            params_str = ""
            if self.log_request_body and hasattr(message, "params") and message.params is not None:
                params_str = f": {self._format_params(message.params)}"

            log_msg = self.request_log_format.format(
                method=message.method, params_str=params_str, id=getattr(message, "id", "none")
            )

            logger.log(self.log_level, log_msg)
        except Exception:
            logger.exception("Error logging request")

    def _log_response(self, message: MCPMessage) -> None:
        """
        Log an MCP response message.
        """
        if not isinstance(message, (JSONRPCResponse, JSONRPCError)):
            logger.debug(f"Ignoring non-response message: {type(message).__name__}")
            return

        try:
            # Calculate request duration if we have the start time
            duration_ms = 0
            request = getattr(message, "_request", None)
            if request:
                start_time = getattr(request, "_mcp_start_time", None)
                if start_time:
                    duration_ms = (time.time() - start_time) * 1000
            else:
                start_time = getattr(message, "_mcp_start_time", None)
                if start_time:
                    duration_ms = (time.time() - start_time) * 1000

            # Skip if below minimum duration threshold
            if self.min_duration_to_log_ms > 0 and duration_ms < self.min_duration_to_log_ms:
                return

            # Handle error responses
            if hasattr(message, "error") and message.error is not None:
                if self.log_errors:
                    error_msg = self.error_log_format.format(
                        method=getattr(message, "method", "unknown"),
                        error=getattr(message.error, "message", str(message.error)),
                        id=getattr(message, "id", "none"),
                    )
                    logger.error(error_msg)
                return

            # Log successful response
            result_str = ""
            if self.log_response_body and hasattr(message, "result"):
                result_str = f": {self._format_result(message.result)}"

            log_msg = self.response_log_format.format(
                method=getattr(message, "method", "unknown"),
                duration=duration_ms,
                id=getattr(message, "id", "none"),
                result_str=result_str,
            )

            logger.log(self.log_level, log_msg)
        except Exception:
            logger.exception("Error logging response")

    def _format_params(self, params: Any) -> str:
        """
        Format parameters for logging.
        """
        try:
            if isinstance(params, dict):
                # Handle common MCP params specially
                if "name" in params and "arguments" in params:
                    return f"{params['name']}({json.dumps(params.get('arguments', {}))})"
                return json.dumps(params)
            return str(params)
        except Exception:
            logger.debug(f"Error formatting params {params!s}")
            return str(params)

    def _format_result(self, result: Any) -> str:
        """
        Format result for logging.
        """
        try:
            if isinstance(result, dict):
                return json.dumps(result)
            return str(result)
        except Exception as e:
            logger.debug(f"Error formatting result {e!s}")
            return str(result)


def create_mcp_logging_middleware(**config: Any) -> MCPLoggingMiddleware:
    """
    Create an MCP logging middleware with the given configuration.

    Args:
        **config: Configuration options.

    Returns:
        An MCPLoggingMiddleware instance.
    """
    return MCPLoggingMiddleware(
        log_level=config.get("log_level", "INFO"),
        log_request_body=config.get("log_request_body", False),
        log_response_body=config.get("log_response_body", False),
        log_errors=config.get("log_errors", True),
        min_duration_to_log_ms=config.get("min_duration_to_log_ms", 0),
        stdio_mode=config.get("stdio_mode", False),
    )



from typing import Annotated

from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Spotify

from arcade_spotify.tools.models import SearchType
from arcade_spotify.tools.utils import (
    get_url,
    send_spotify_request,
)


@tool(requires_auth=Spotify())
async def search(
    context: ToolContext,
    q: Annotated[str, "The search query"],
    types: Annotated[list[SearchType], "The types of results to return"],
    limit: Annotated[int, "The maximum number of results to return"] = 1,
) -> Annotated[dict, "A list of artists matching the search query"]:
    """Search Spotify catalog information

    Explanation of the q parameter:
        You can narrow down your search using field filters.
        Available filters are album, artist, track, year, upc, tag:hipster, tag:new, isrc, and
        genre. Each field filter only applies to certain result types.

        The artist and year filters can be used while searching albums, artists and tracks.
        You can filter on a single year or a range (e.g. 1955-1960).
        The album filter can be used while searching albums and tracks.
        The genre filter can be used while searching artists and tracks.
        The isrc and track filters can be used while searching tracks.
        The upc, tag:new and tag:hipster filters can only be used while searching albums.
        The tag:new filter will return albums released in the past two weeks and tag:hipster
        can be used to return only albums with the lowest 10% popularity.

        Example: q="remaster track:Doxy artist:Miles Davis"
    """

    url = get_url("search", q=q)

    response = await send_spotify_request(
        context, "GET", url, params={"q": q, "type": ",".join(types), "limit": limit}
    )
    response.raise_for_status()
    return dict(response.json())



from typing import Any

import httpx
from arcade_tdk.errors import ToolExecutionError

from arcade_github.tools.constants import ENDPOINTS, GITHUB_API_BASE_URL


def handle_github_response(response: httpx.Response, url: str) -> None:
    """
    Handle GitHub API response and raise appropriate exceptions for non-200 status codes.

    :param response: The response object from the GitHub API
    :param url: The URL of the API endpoint
    :raises ToolExecutionError: If the response status code is not 200
    """
    if 200 <= response.status_code < 300:
        return

    error_messages = {
        301: "Moved permanently. The repository has moved.",
        304: "Not modified. The requested resource hasn't been modified since the last request.",
        403: "Forbidden. You do not have access to this resource.",
        404: "Resource not found. The requested resource does not exist.",
        410: "Gone. The requested resource is no longer available.",
        422: "Validation failed or the endpoint has been spammed.",
        503: "Service unavailable. The server is temporarily unable to handle the request.",
    }

    error_message = error_messages.get(
        response.status_code, f"Failed to process request. Status code: {response.status_code}"
    )

    raise ToolExecutionError(f"Error accessing '{url}': {error_message}")


def get_github_json_headers(token: str | None) -> dict:
    """
    Generate common headers for GitHub API requests.

    :param token: The authorization token
    :return: A dictionary of headers
    """
    token = token or ""
    return {
        "Accept": "application/vnd.github+json",
        "Authorization": f"Bearer {token}",
        "X-GitHub-Api-Version": "2022-11-28",
    }


def get_github_diff_headers(token: str | None) -> dict:
    """
    Generate headers for GitHub API requests for diff content.

    :param token: The authorization token
    :return: A dictionary of headers
    """
    token = token or ""
    return {
        "Accept": "application/vnd.github.diff",
        "Authorization": f"Bearer {token}",
        "X-GitHub-Api-Version": "2022-11-28",
    }


def remove_none_values(params: dict) -> dict:
    """
    Remove None values from a dictionary.

    :param params: The dictionary to clean
    :return: A new dictionary with None values removed
    """
    return {k: v for k, v in params.items() if v is not None}


def get_url(endpoint: str, **kwargs: Any) -> str:
    """
    Get the full URL for a given endpoint.

    :param endpoint: The endpoint key from ENDPOINTS
    :param kwargs: The parameters to format the URL with
    :return: The full URL
    """
    return f"{GITHUB_API_BASE_URL}{ENDPOINTS[endpoint].format(**kwargs)}"



from typing import Any

import httpx
from arcade_tdk import ToolContext

from arcade_notion_toolkit.constants import ENDPOINTS, NOTION_API_URL, UNTITLED_TITLE


def get_url(endpoint: str, **kwargs: Any) -> str:
    """
    Constructs the full URL for a specified notion endpoint.

    Args:
        endpoint (str): The endpoint key from ENDPOINTS.
        **kwargs: Additional parameters to format the URL.

    Returns:
        str: The complete URL for the specified endpoint.
    """
    return f"{NOTION_API_URL}{ENDPOINTS[endpoint].format(**kwargs)}"


def get_headers(context: ToolContext) -> dict[str, str]:
    """
    Retrieves the headers for a given context.

    Args:
        context (ToolContext): The context containing authorization and other information.

    Returns:
        dict[str, str]: A dictionary containing the headers for the Notion API request.
    """
    return {
        "Authorization": context.get_auth_token_or_empty(),
        "Content-Type": "application/json",
        "Notion-Version": "2022-06-28",
    }


def remove_none_values(payload: dict[str, Any]) -> dict[str, Any]:
    """
    Removes all keys with None values from a dictionary.

    Args:
        payload (dict[str, Any]): The dictionary to remove None values from.

    Returns:
        dict[str, Any]: A dictionary with all None values removed.
    """
    return {k: v for k, v in payload.items() if v is not None}


def extract_title(item: dict) -> str:
    """
    Extracts a human-readable title from a page or database, or a block if possible.

    Args:
        item (dict): The item to extract the title from.

    Returns:
        str: The human-readable title of the item.
    """
    properties: dict = item.get("properties", {})
    # Case 1: Extract title from a database object.
    if item["object"] == "database" and "title" in item:
        return "".join([t.get("plain_text", "") for t in item.get("title", [])])

    # Case 2: Extract title from a page object that is parented by the workspace or a page
    if item["object"] == "page" and "title" in properties:
        return "".join([t["plain_text"] for t in properties["title"].get("title", [])])

    # Case 3: Extract title from a page object that is parented a database
    elif item["object"] == "page":
        for prop in properties.values():
            if isinstance(prop, dict) and prop.get("type") == "title":
                return "".join([t.get("plain_text", "") for t in prop.get("title", [])])

    # Case 4: Extract title from a child page block object
    if item.get("object") == "block":
        block_type = item.get("type")
        if block_type == "child_page":
            title: str = item.get("child_page", {}).get("title", UNTITLED_TITLE)
            return title
        # For text-based blocks, try extracting rich_text.
        if block_type in ["paragraph", "heading_1", "heading_2", "heading_3"]:
            rich_text = item.get(block_type, {}).get("rich_text", [])
            return "".join([t.get("plain_text", "") for t in rich_text]) or block_type

    return UNTITLED_TITLE


def simplify_search_result(item: dict) -> dict:
    """
    Simplifies a 'search by title' result from the Notion API.
    Takes a page object or database object and extracts only the necessary data.

    Args:
        item (dict): The search result to simplify.

    Returns:
        dict: A simplified search result
    """
    title = extract_title(item)

    return {
        "id": item.get("id"),
        "object": item.get("object"),
        "parent": item.get("parent"),
        "created_time": item.get("created_time"),
        "last_edited_time": item.get("last_edited_time"),
        "title": title,
        "url": item.get("url"),
        "public_url": item.get("public_url"),
    }


async def get_next_page(
    client: httpx.AsyncClient,
    url: str,
    headers: dict,
    params: dict | None = None,
    cursor: str | None = None,
) -> tuple[dict, bool, str]:
    """
    Retrieves the next page of results from a Notion API endpoint.
    This is a helper function that is useful when paginating through Notion API responses.

    Args:
        client (httpx.AsyncClient): The HTTP client to use for the request.
        url (str): The URL of the endpoint to request.
        headers (dict): The headers to use for the request.
        params (dict | None): The parameters to use for the request.
        cursor (str | None): The cursor to use for the request.

    Returns:
        tuple[dict, bool, str]: A tuple containing the results, a boolean indicating if there is a
        next page, and the next cursor.
    """
    params = params or {}
    if cursor:
        params["start_cursor"] = cursor
    elif "start_cursor" in params:
        del params["start_cursor"]

    response = await client.get(url, headers=headers, params=params)
    response.raise_for_status()
    data = response.json()
    return data, data.get("has_more", False), data.get("next_cursor")


async def get_page_url(context: ToolContext, page_id: str) -> str:
    """
    Retrieves the URL of a page from the Notion API.

    Args:
        context (ToolContext): The context containing authorization and other information.
        page_id (str): The ID of the page to get the URL of.

    Returns:
        str: The URL of the page or an empty string if the page's metadata cannot be retrieved.
    """
    url = get_url("retrieve_a_page", page_id=page_id)
    headers = get_headers(context)
    async with httpx.AsyncClient() as client:
        response = await client.get(url, headers=headers)
        if response.status_code != 200:
            return ""
        data = response.json()
        return data.get("url", "")  # type: ignore[no-any-return]


def build_workspace_structure(items: list[dict[str, Any]]) -> dict[str, list]:
    """Build a tree structure from a flat list of Notion objects.

    Args:
        items (list[dict[str, Any]]): A list of Notion objects.

    Returns:
        dict[str, list]: A tree structure of the workspace.
    """
    # For each item, we initialize a children list and then attach it
    # under its parent if one exists.
    nodes = {}
    for item in items:
        node = item.copy()
        node["children"] = []
        nodes[node["id"]] = node

    roots = []
    for node in nodes.values():
        parent = node.get("parent", {})
        parent_type = parent.get("type")
        if parent_type == "workspace":
            # No parent beyond workspace i.e., the node is a root.
            roots.append(node)
        elif parent_type == "page_id":
            parent_id = parent.get("page_id")
            if parent_id and parent_id in nodes:
                nodes[parent_id]["children"].append(node)
            else:
                roots.append(node)
        elif parent_type == "database_id":
            parent_id = parent.get("database_id")
            if parent_id and parent_id in nodes:
                nodes[parent_id]["children"].append(node)
            else:
                roots.append(node)
        else:
            # Fallback: if parent's type is missing or unrecognized, then treat as root.
            roots.append(node)

    def prune_node(node: dict) -> dict:
        """Get rid of all of the unnecessary fields in a node"""
        pruned_node = {
            "id": node["id"],
            "title": node["title"],
            "type": node["object"],
            "url": node["url"],
        }
        if node.get("children"):
            pruned_node["children"] = [prune_node(child) for child in node["children"]]

        return pruned_node

    return {"workspace": [prune_node(root) for root in roots]}



from unittest.mock import AsyncMock, patch

import pytest
from arcade_tdk.errors import ToolExecutionError
from googleapiclient.errors import HttpError

from arcade_google_docs.enum import Corpora, DocumentFormat, OrderBy
from arcade_google_docs.templates import optional_file_picker_instructions_template
from arcade_google_docs.tools import (
    search_and_retrieve_documents,
    search_documents,
)
from arcade_google_docs.utils import build_drive_service


@pytest.fixture
def mock_context():
    context = AsyncMock()
    context.authorization.token = "mock_token"  # noqa: S105
    context.get_metadata.side_effect = lambda key: {
        "client_id": "123456789-abcdefg.apps.googleusercontent.com",
        "coordinator_url": "https://coordinator.example.com",
    }.get(key.value if hasattr(key, "value") else key)
    return context


@pytest.fixture
def mock_service():
    with patch(
        "arcade_google_docs.tools.search." + build_drive_service.__name__
    ) as mock_build_service:
        yield mock_build_service.return_value


@pytest.mark.asyncio
async def test_search_documents_success(mock_context, mock_service):
    # Mock the service.files().list().execute() method
    mock_service.files.return_value.list.return_value.execute.side_effect = [
        {
            "files": [
                {"id": "file1", "name": "Document 1"},
                {"id": "file2", "name": "Document 2"},
            ],
            "nextPageToken": None,
        }
    ]

    # Mock the generate_google_file_picker_url function
    with patch(
        "arcade_google_docs.tools.search.generate_google_file_picker_url"
    ) as mock_file_picker:
        mock_file_picker.return_value = {
            "url": "https://coordinator.example.com/google/drive_picker?config=test_config",
            "llm_instructions": optional_file_picker_instructions_template.format(
                url="https://coordinator.example.com/google/drive_picker?config=test_config"
            ),
        }

        result = await search_documents(mock_context, limit=2)

    assert result["documents_count"] == 2
    assert len(result["documents"]) == 2
    assert result["documents"][0]["id"] == "file1"
    assert result["documents"][1]["id"] == "file2"


@pytest.mark.asyncio
async def test_search_documents_pagination(mock_context, mock_service):
    # Simulate multiple pages
    mock_service.files.return_value.list.return_value.execute.side_effect = [
        {
            "files": [{"id": f"file{i}", "name": f"Document {i}"} for i in range(1, 11)],
            "nextPageToken": "token1",
        },
        {
            "files": [{"id": f"file{i}", "name": f"Document {i}"} for i in range(11, 21)],
            "nextPageToken": None,
        },
    ]

    # Mock the generate_google_file_picker_url function
    with patch(
        "arcade_google_docs.tools.search.generate_google_file_picker_url"
    ) as mock_file_picker:
        mock_file_picker.return_value = {
            "url": "https://coordinator.example.com/google/drive_picker?config=test_config",
            "llm_instructions": optional_file_picker_instructions_template.format(
                url="https://coordinator.example.com/google/drive_picker?config=test_config"
            ),
        }

        result = await search_documents(mock_context, limit=15)

    assert result["documents_count"] == 15
    assert len(result["documents"]) == 15
    assert result["documents"][0]["id"] == "file1"
    assert result["documents"][-1]["id"] == "file15"


@pytest.mark.asyncio
async def test_search_documents_http_error(mock_context, mock_service):
    # Simulate HttpError
    mock_service.files.return_value.list.return_value.execute.side_effect = HttpError(
        resp=AsyncMock(status=403), content=b'{"error": {"message": "Forbidden"}}'
    )

    with pytest.raises(
        ToolExecutionError, match=f"Error in execution of {search_documents.__tool_name__}"
    ):
        await search_documents(mock_context)


@pytest.mark.asyncio
async def test_search_documents_unexpected_error(mock_context, mock_service):
    # Simulate unexpected exception
    mock_service.files.return_value.list.return_value.execute.side_effect = Exception(
        "Unexpected error"
    )

    with pytest.raises(
        ToolExecutionError, match=f"Error in execution of {search_documents.__tool_name__}"
    ):
        await search_documents(mock_context)


@pytest.mark.asyncio
async def test_search_documents_in_organization_domains(mock_context, mock_service):
    # Mock the service.files().list().execute() method
    mock_service.files.return_value.list.return_value.execute.side_effect = [
        {
            "files": [
                {"id": "file1", "name": "Document 1"},
            ],
            "nextPageToken": None,
        }
    ]

    # Mock the generate_google_file_picker_url function
    with patch(
        "arcade_google_docs.tools.search.generate_google_file_picker_url"
    ) as mock_file_picker:
        mock_file_picker.return_value = {
            "url": "https://coordinator.example.com/google/drive_picker?config=test_config",
            "llm_instructions": optional_file_picker_instructions_template.format(
                url="https://coordinator.example.com/google/drive_picker?config=test_config"
            ),
        }

        result = await search_documents(
            mock_context,
            order_by=OrderBy.MODIFIED_TIME_DESC,
            include_shared_drives=False,
            include_organization_domain_documents=True,
            limit=1,
        )

    assert result["documents_count"] == 1
    mock_service.files.return_value.list.assert_called_with(
        q="(mimeType = 'application/vnd.google-apps.document' and trashed = false)",
        corpora=Corpora.DOMAIN.value,
        pageSize=1,
        orderBy=OrderBy.MODIFIED_TIME_DESC.value,
        includeItemsFromAllDrives="true",
        supportsAllDrives="true",
    )


@pytest.mark.asyncio
@patch("arcade_google_docs.tools.search.search_documents")
@patch("arcade_google_docs.tools.search.get_document_by_id")
async def test_search_and_retrieve_documents_in_markdown_format(
    mock_get_document_by_id,
    mock_search_documents,
    mock_context,
    sample_document_and_expected_formats,
):
    (sample_document, expected_markdown, _) = sample_document_and_expected_formats
    mock_search_documents.return_value = {
        "documents_count": 1,
        "documents": [{"id": sample_document["documentId"], "title": sample_document["title"]}],
    }
    mock_get_document_by_id.return_value = sample_document

    # Mock the generate_google_file_picker_url function
    with patch(
        "arcade_google_docs.tools.search.generate_google_file_picker_url"
    ) as mock_file_picker:
        mock_file_picker.return_value = {
            "url": "https://coordinator.example.com/google/drive_picker?config=test_config",
            "llm_instructions": optional_file_picker_instructions_template.format(
                url="https://coordinator.example.com/google/drive_picker?config=test_config"
            ),
        }

        result = await search_and_retrieve_documents(
            mock_context,
            document_contains=[sample_document["title"]],
            return_format=DocumentFormat.MARKDOWN,
        )

    assert result["documents_count"] == 1
    assert result["documents"][0] == expected_markdown


@pytest.mark.asyncio
@patch("arcade_google_docs.tools.search.search_documents")
@patch("arcade_google_docs.tools.search.get_document_by_id")
async def test_search_and_retrieve_documents_in_html_format(
    mock_get_document_by_id,
    mock_search_documents,
    mock_context,
    sample_document_and_expected_formats,
):
    (sample_document, _, expected_html) = sample_document_and_expected_formats
    mock_search_documents.return_value = {
        "documents_count": 1,
        "documents": [{"id": sample_document["documentId"], "title": sample_document["title"]}],
    }
    mock_get_document_by_id.return_value = sample_document

    # Mock the generate_google_file_picker_url function
    with patch(
        "arcade_google_docs.tools.search.generate_google_file_picker_url"
    ) as mock_file_picker:
        mock_file_picker.return_value = {
            "url": "https://coordinator.example.com/google/drive_picker?config=test_config",
            "llm_instructions": optional_file_picker_instructions_template.format(
                url="https://coordinator.example.com/google/drive_picker?config=test_config"
            ),
        }

        result = await search_and_retrieve_documents(
            mock_context,
            document_contains=[sample_document["title"]],
            return_format=DocumentFormat.HTML,
        )

    assert result["documents_count"] == 1
    assert result["documents"][0] == expected_html


@pytest.mark.asyncio
@patch("arcade_google_docs.tools.search.search_documents")
@patch("arcade_google_docs.tools.search.get_document_by_id")
async def test_search_and_retrieve_documents_in_google_json_format(
    mock_get_document_by_id,
    mock_search_documents,
    mock_context,
    sample_document_and_expected_formats,
):
    (sample_document, _, _) = sample_document_and_expected_formats
    mock_search_documents.return_value = {
        "documents_count": 1,
        "documents": [{"id": sample_document["documentId"], "title": sample_document["title"]}],
    }
    mock_get_document_by_id.return_value = sample_document

    # Mock the generate_google_file_picker_url function
    with patch(
        "arcade_google_docs.tools.search.generate_google_file_picker_url"
    ) as mock_file_picker:
        mock_file_picker.return_value = {
            "url": "https://coordinator.example.com/google/drive_picker?config=test_config",
            "llm_instructions": optional_file_picker_instructions_template.format(
                url="https://coordinator.example.com/google/drive_picker?config=test_config"
            ),
        }

        result = await search_and_retrieve_documents(
            mock_context,
            document_contains=[sample_document["title"]],
            return_format=DocumentFormat.GOOGLE_API_JSON,
        )

    assert result["documents_count"] == 1
    assert result["documents"][0] == sample_document



from arcade_tdk.errors import ToolExecutionError


class JiraToolExecutionError(ToolExecutionError):
    pass


class NotFoundError(JiraToolExecutionError):
    pass


class MultipleItemsFoundError(JiraToolExecutionError):
    pass



from typing import Annotated

from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Zoom

from arcade_zoom.tools.utils import _handle_zoom_api_error, _send_zoom_request


@tool(
    requires_auth=Zoom(
        scopes=["meeting:read:list_upcoming_meetings"],
    )
)
async def list_upcoming_meetings(
    context: ToolContext,
    user_id: Annotated[
        str | None,
        "The user's user ID or email address. Defaults to 'me' for the current user.",
    ] = "me",
) -> Annotated[dict, "List of upcoming meetings within the next 24 hours"]:
    """List a Zoom user's upcoming meetings within the next 24 hours."""
    endpoint = f"/users/{user_id}/upcoming_meetings"

    response = await _send_zoom_request(context, "GET", endpoint)

    if not (200 <= response.status_code < 300):
        _handle_zoom_api_error(response)

    response_json = response.json()
    return dict(response_json)


@tool(
    requires_auth=Zoom(
        scopes=["meeting:read:invitation"],
    )
)
async def get_meeting_invitation(
    context: ToolContext,
    meeting_id: Annotated[
        str,
        "The meeting's numeric ID (as a string).",
    ],
) -> Annotated[dict, "Meeting invitation string"]:
    """Retrieve the invitation note for a specific Zoom meeting."""
    endpoint = f"/meetings/{meeting_id}/invitation"

    response = await _send_zoom_request(context, "GET", endpoint)

    if not (200 <= response.status_code < 300):
        _handle_zoom_api_error(response)

    response_json = response.json()
    return dict(response_json)
